from regulations.generator.sidebar.base import SidebarBase


class DiffHelp(SidebarBase):
    """Help info specific to Diffs"""
    shorthand = 'diff_help'

    def context(self, http_client, request):
        return {}



#!/usr/bin/env python
# vim: set fileencoding=utf-8
from __future__ import unicode_literals

import re
from itertools import takewhile

from six.moves import filter, filterfalse

from regulations.generator import node_types
from regulations.generator.layers.layers_applier import LayersApplier
from regulations.generator.layers.internal_citation import (
    InternalCitationLayer)
from regulations.apps import RegulationsConfig
from .link_flattener import flatten_links


class HTMLBuilder(object):
    # @todo simplify this constructor
    def __init__(self, inline_applier, p_applier, search_applier,
                 diff_applier=None, id_prefix=None, index_prefix=None):
        self.tree = None
        self.inline_applier = inline_applier
        self.p_applier = p_applier
        self.search_applier = search_applier
        self.diff_applier = diff_applier
        self.id_prefix = id_prefix or []
        self.index_prefix = index_prefix or []
        self.preprocess()

    def preprocess(self):
        """Noop. Hook for subclasses"""
        pass

    def generate_html(self):
        if self.diff_applier:
            self.diff_applier.tree_changes(self.tree)
        for layer in self.p_applier.layers.values():
            if hasattr(layer, 'preprocess_root'):   # @todo - remove
                layer.preprocess_root(self.tree)
        self.process_node(self.tree, indexes=self.index_prefix)

    def list_level(self, parts, node_type):
        return len(parts) - 2

    def process_node_title(self, node):
        if 'title' in node:
            node['header'] = node['title']
            if self.diff_applier:
                node['header'] = self.diff_applier.apply_diff(
                    node['header'], node['label_id'], component='title')

    @staticmethod
    def is_collapsed(node):
        """A "collapsed" paragraph is one which has no text, immediately
        starting a subparagraph. For example:
        (a)(1) Some text    <- (a) is collapsed
        (a) Some text - (1) Other text   <- (a) is not collapsed
        """
        marker = '({})'.format(node['label'][-1])
        text_without_marker = node['text'].replace(marker, '')
        return node['text'].strip() and not text_without_marker.strip()

    def process_node(self, node, indexes=None):
        """Every node passes through this function on the way to being
        rendered. Importantly, this adds the `marked_up` field, which contains
        the HTML version of node's text (after applying all relevant
        layers) and the `template_name` field, which defines how this node
        should be rendered."""

        node['label_id'] = '-'.join(node['label'])
        self.process_node_title(node)
        node['is_collapsed'] = self.is_collapsed(node)

        node['indexes'] = indexes or []
        node['html_label'] = node_types.to_markup_id(node['label'])
        node['markup_id'] = "-".join(node['html_label'])
        node['full_id'] = "-".join(self.id_prefix + node['html_label'])
        node['tree_level'] = len(node['label']) - 1
        node['human_label'] = self.human_label(node)

        node['list_level'] = self.list_level(node['label'], node['node_type'])

        if len(node['text']):
            inline_elements = self.inline_applier.get_layer_pairs(
                node['label_id'], node['text'])
            search_elements = self.search_applier.get_layer_pairs(
                node['label_id'])

            if self.diff_applier:
                node['marked_up'] = self.diff_applier.apply_diff(
                    node['text'], node['label_id'])

            layers_applier = LayersApplier()
            layers_applier.enqueue_from_list(inline_elements)
            layers_applier.enqueue_from_list(search_elements)

            node['marked_up'] = layers_applier.apply_layers(
                node.get('marked_up', node['text']))
            node['marked_up'] = flatten_links(node['marked_up'])

        node = self.p_applier.apply_layers(node)

        node['template_name'] = RegulationsConfig.custom_tpls.get(
            node['label_id'],
            RegulationsConfig.node_type_tpls[node['node_type'].lower()])

        for idx, child in enumerate(node['children']):
            self.process_node(child, indexes=node['indexes'] + [idx])

    @staticmethod
    def human_label(node):
        """Derive a human-readable description for this node"""
        return '-'.join(node['label'])      # Default


class CFRHTMLBuilder(HTMLBuilder):
    SECTION_NUMBER_REGEX = re.compile(r'(§+)\s+')

    @classmethod
    def section_space(cls, text):
        """ After a section sign, insert a non-breaking space. """
        return cls.SECTION_NUMBER_REGEX.sub(r'\1&nbsp;', text)

    def list_level(self, parts, node_type):
        """ Return the list level and the list type. Overrides"""
        if node_type == node_types.INTERP:
            prefix_length = parts.index('Interp')+1
        elif node_type == node_types.APPENDIX:
            prefix_length = 3
        else:
            prefix_length = 2

        if len(parts) > prefix_length:
            return len(parts) - prefix_length
        return 0

    def process_node_title(self, node):
        """Add space to header. Overrides"""
        super(CFRHTMLBuilder, self).process_node_title(node)
        if 'header' in node:
            node['header'] = self.section_space(node['header'])

    def process_node(self, node, indexes=None):
        """Overrides with custom, additional processing"""
        super(CFRHTMLBuilder, self).process_node(node, indexes=indexes)
        if 'marked_up' in node:
            node['marked_up'] = self.section_space(node['marked_up'])

        if 'TOC' in node:
            for l in node['TOC']:
                l['label'] = self.section_space(l['label'])

        if 'interp' in node and 'markup' in node['interp']:
            node['interp']['markup'] = self.section_space(
                node['interp']['markup'])

        if node['node_type'] == node_types.INTERP:
            self.modify_interp_node(node)

    def modify_interp_node(self, node):
        """Add extra fields which only exist on interp nodes"""
        #   ['105', '22', 'Interp'] => section header
        node['section_header'] = len(node['label']) == 3

        def is_header(child):
            return child['label'][-1] == 'Interp'

        node['header_children'] = list(filter(is_header, node['children']))
        node['par_children'] = list(filterfalse(is_header, node['children']))
        if 'header' in node:
            node['header_markup'] = node['header']
            citation = list(takewhile(lambda p: p != 'Interp',
                                      node['label']))
            icl = self.inline_applier.layers.get(
                InternalCitationLayer.shorthand)
            if icl and len(citation) > 2:
                text = '%s(%s)' % (citation[1], ')('.join(citation[2:]))
                node['header_markup'] = node['header_markup'].replace(
                    text, icl.render_url(citation, text))

    @staticmethod
    def human_label(node):
        """Derive a human-readable description for this node. Override"""
        return node_types.label_to_text(node['label'], include_marker=True)


class PreambleHTMLBuilder(HTMLBuilder):
    @staticmethod
    def human_label(node):
        """Derive a human-readable description for this node. Override"""
        is_markerless = node_types.MARKERLESS_REGEX.match
        prefix = list(takewhile(lambda l: not is_markerless(l),
                                node['label']))
        if 'intro' in prefix:
            title = node.get('title', '').rstrip(':')
            return 'Intro: ' + title
        elif len(prefix) > 1:
            label = 'Section ' + '.'.join(prefix[1:])
            count = len(node['label']) - len(prefix)
            if count and node.get('indexes'):
                paragraphs = '.'.join(
                    str(idx + 1)
                    for idx in node['indexes'][-count:]
                )
                label += ' Paragraph ' + paragraphs
            return label
        else:
            return 'FR #' + prefix[0]

    def process_node(self, node, indexes=None):
        """Overrides with custom, additional processing"""
        super(PreambleHTMLBuilder, self).process_node(node, indexes=indexes)
        node['accepts_comments'] = True
        node['comments_calledout'] = bool(node.get('title'))

        def not_markerless(l):
            return not node_types.MARKERLESS_REGEX.match(l)

        markers = takewhile(not_markerless, node['label'][:4])
        node['toc_id'] = '-'.join(self.id_prefix + list(markers))


class CFRChangeHTMLBuilder(CFRHTMLBuilder):
    """Generated HTML specifically related to changing CFR data, as displayed
    in a notice. This assumes self.diff_applier is set"""
    def process_node(self, node, indexes=None):
        """Overrides with custom, additional processing"""
        super(CFRHTMLBuilder, self).process_node(node, indexes=indexes)
        label_id = '-'.join(node['label'])
        node['toc_id'] = '-'.join(self.id_prefix + node['label'][:2])
        node['accepts_comments'] = label_id in self.diff_applier.diff
        node['comments_calledout'] = label_id in self.diff_applier.diff

        has_diff = label_id in self.diff_applier.diff
        on_diff_path = tuple(node['label']) in self.diff_paths
        if has_diff:
            node['stars_collapse'] = 'none'
        elif on_diff_path:
            node['stars_collapse'] = 'inline'
        else:
            node['stars_collapse'] = 'full'

    def preprocess(self):
        """Pre-generate all of the "paths" associated with diffs; if there's a
        diff for 111-22-c-4-v, we'd capture
        (111,) (111, 22) (111, 22, c) (111, 22, c, 4) and (111, 22, c, 4, v)"""
        self.diff_paths = set()
        for label_id in self.diff_applier.diff:
            label_parts = tuple(label_id.split('-'))
            path = tuple()
            for part in label_parts:
                path = path + (part,)
                self.diff_paths.add(path)



import os
from os.path import join, abspath, dirname
import tempfile

from distutils.spawn import find_executable
from django.utils.crypto import get_random_string


here = lambda *x: join(abspath(dirname(__file__)), *x)
PROJECT_ROOT = here("..", "..")
root = lambda *x: join(abspath(PROJECT_ROOT), *x)

DEBUG = True

ADMINS = (
    # ('Your Name', 'your_email@example.com'),
)

MANAGERS = ADMINS

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'regsite.db'
    }
}

# Hosts/domain names that are valid for this site; required if DEBUG is False
# See https://docs.djangoproject.com/en/1.5/ref/settings/#allowed-hosts
ALLOWED_HOSTS = []

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# In a Windows environment this must be set to your system time zone.
TIME_ZONE = 'America/New_York'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# If you set this to False, Django will not format dates, numbers and
# calendars according to the current locale.
USE_L10N = True

# If you set this to False, Django will not use timezone-aware datetimes.
USE_TZ = True

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/var/www/example.com/media/"
MEDIA_ROOT = ''

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash.
# Examples: "http://example.com/media/", "http://media.example.com/"
MEDIA_URL = ''

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/var/www/example.com/static/"
if 'TMPDIR' in os.environ:
    STATIC_ROOT = os.environ['TMPDIR'] + '/static/'
else:
    STATIC_ROOT = ''

# URL prefix for static files.
# Example: "http://example.com/static/", "http://static.example.com/"
STATIC_URL = '/static/'

# Additional locations of static files
STATICFILES_DIRS = (
    # Put strings here, like "/home/html/static" or "C:/www/django/static".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
)

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)

SECRET_KEY = os.environ.get('DJANGO_SECRET_KEY', get_random_string(50))

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "OPTIONS": {
            "context_processors": (
                # "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
                "django.core.context_processors.request",
                "regulations.context.eregs_globals",
            ),
            # List of callables that know how to import templates from various
            # sources.
            "loaders": [
                ('django.template.loaders.cached.Loader', (
                    'django.template.loaders.filesystem.Loader',
                    'django.template.loaders.app_directories.Loader'))
            ],
        }
    },
]


# Note order:
# https://docs.djangoproject.com/en/1.8/topics/cache/#the-per-site-cache
MIDDLEWARE_CLASSES = (
    'django.middleware.cache.UpdateCacheMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.middleware.cache.FetchFromCacheMiddleware',
)

ROOT_URLCONF = 'regulations.urls'

INSTALLED_APPS = (
    'django.contrib.staticfiles',
    'regulations.apps.RegulationsConfig',
)

# eregs specific settings

# The base URL for the API that we use to access layers and the regulation.
API_BASE = os.environ.get('EREGS_API_BASE', '')

DATE_FORMAT = 'n/j/Y'

# Analytics settings

ANALYTICS = {
    'GOOGLE': {
       'GTM_SITE_ID': '',
       'GA_SITE_ID': '',
    },
    'DAP': {
        'AGENCY': '',
        'SUBAGENCY': '',
    },
}

#   Use the 'source' directory; useful for development
JS_DEBUG = False

# Django by default tries to setup a databse when running tests. We don't have
# a database, so we override the default test runner.
TEST_RUNNER = 'regulations.tests.runner.DatabaselessTestRunner'

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.filebased.FileBasedCache',
        'LOCATION': tempfile.mkdtemp('eregs_cache'),
    },
    'eregs_longterm_cache': {
        'BACKEND': 'django.core.cache.backends.filebased.FileBasedCache',
        'LOCATION': tempfile.mkdtemp('eregs_longterm_cache'),
        'TIMEOUT': 60*60*24*15,     # 15 days
        'OPTIONS': {
            'MAX_ENTRIES': 10000,
        },
    },
    'api_cache': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'api_cache_memory',
        'TIMEOUT': 3600,
        'OPTIONS': {
            'MAX_ENTRIES': 1000,
        },
    },
    'regs_gov_cache': {
        'BACKEND': 'django.core.cache.backends.dummy.DummyCache',
    }
}

CACHE_MIDDLEWARE_ALIAS = 'default'
CACHE_MIDDLEWARE_KEY_PREFIX = 'eregs'
CACHE_MIDDLEWARE_SECONDS = 600
# A sample logging configuration. The only tangible logging
# performed by this configuration is to send an email to
# the site admins on every HTTP 500 error when DEBUG=False.
# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        }
    },
    'handlers': {
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler'
        }
    },
    'loggers': {
        'django.request': {
            'handlers': ['mail_admins'],
            'level': 'ERROR',
            'propagate': True,
        },
    }
}


# Where should we look for data?
DATA_LAYERS = (
    'regulations.generator.layers.defined.DefinedLayer',
    'regulations.generator.layers.definitions.DefinitionsLayer',
    'regulations.generator.layers.external_citation.ExternalCitationLayer',
    'regulations.generator.layers.footnotes.FootnotesLayer',
    'regulations.generator.layers.formatting.FormattingLayer',
    'regulations.generator.layers.internal_citation.InternalCitationLayer',
    # Should likely be moved to a CFPB-specific module
    'regulations.generator.layers.interpretations.InterpretationsLayer',
    'regulations.generator.layers.key_terms.KeyTermsLayer',
    'regulations.generator.layers.meta.MetaLayer',
    'regulations.generator.layers.paragraph_markers.MarkerHidingLayer',
    'regulations.generator.layers.paragraph_markers.MarkerInfoLayer',
    'regulations.generator.layers.toc_applier.TableOfContentsLayer',
    'regulations.generator.layers.graphics.GraphicsLayer',
)

SIDEBARS = (
    'regulations.generator.sidebar.analyses.Analyses',
    'regulations.generator.sidebar.help.Help',
)

ATTACHMENT_BUCKET = os.getenv('S3_BUCKET')
ATTACHMENT_ACCESS_KEY_ID = os.getenv('S3_ACCESS_KEY_ID')
ATTACHMENT_SECRET_ACCESS_KEY = os.getenv('S3_SECRET_ACCESS_KEY')

# regulations.gov restrictions
ATTACHMENT_MAX_SIZE = 1024 * 1024 * 10
ATTACHMENT_PREVIEW_PREFIX = 'preview'
VALID_ATTACHMENT_EXTENSIONS = set([
    "bmp", "doc", "xls", "pdf", "gif", "htm", "html", "jpg", "jpeg",
    "png", "ppt", "rtf", "sgml", "tiff", "tif", "txt", "wpd", "xml",
    "docx", "xlsx", "pptx"])
MAX_ATTACHMENT_COUNT = 10

REGS_GOV_API_URL = os.environ.get('REGS_GOV_API_URL')
REGS_GOV_API_KEY = os.environ.get('REGS_GOV_API_KEY')

COMMENT_DOCUMENT_ID = os.getenv('DOCUMENT_ID')
WKHTMLTOPDF_PATH = os.getenv('WKHTMLTOPDF_PATH',
                             find_executable('wkhtmltopdf'))



# vim: set encoding=utf-8
from unittest import TestCase
from mock import Mock

from regulations.generator.html_builder import (
    CFRChangeHTMLBuilder, CFRHTMLBuilder, HTMLBuilder, PreambleHTMLBuilder)
from regulations.generator.layers.diff_applier import DiffApplier
from regulations.generator.layers.internal_citation import (
    InternalCitationLayer)
from regulations.generator.layers.layers_applier import InlineLayersApplier
from regulations.generator.layers.layers_applier import ParagraphLayersApplier
from regulations.generator.node_types import REGTEXT, APPENDIX, INTERP
from regulations.generator.layers import diff_applier


class HTMLBuilderTest(TestCase):
    def test_process_node_appliers(self):
        node = {
            "text": "Text text text.",
            "children": [],
            "label": ["123", "aaa"],
            'node_type': REGTEXT
        }

        inline = Mock()
        inline.get_layer_pairs.return_value = []
        par = Mock()
        par.apply_layers.return_value = node
        sr = Mock()
        sr.get_layer_pairs.return_value = []

        builder = HTMLBuilder(inline, par, sr)
        builder.process_node(node)

        self.assertTrue(inline.get_layer_pairs.called)
        self.assertEqual("123-aaa",
                         inline.get_layer_pairs.call_args[0][0])
        self.assertEqual("Text text text.",
                         inline.get_layer_pairs.call_args[0][1])

        self.assertTrue(par.apply_layers.called)
        self.assertEqual(node, par.apply_layers.call_args[0][0])

    def test_process_node_header(self):
        builder = HTMLBuilder(None, ParagraphLayersApplier(), None)
        node = {'text': '', 'children': [], 'label': ['99', '22'],
                'node_type': REGTEXT}
        builder.process_node(node)
        self.assertFalse('header' in node)

        node = {'text': '', 'children': [], 'label': ['99', '22'],
                'title': 'Some Title', 'node_type': REGTEXT}
        builder.process_node(node)
        self.assertTrue('header' in node)
        self.assertEqual('Some Title', node['header'])

        node = {'text': '', 'children': [], 'label': ['99', '22'],
                'title': u'§ 22.1 Title', 'node_type': REGTEXT}
        builder.process_node(node)
        self.assertTrue('header' in node)

    def test_process_node_title_diff(self):
        builder = HTMLBuilder(None, None, None)
        diff = {'204': {'title': [('delete', 0, 2), ('insert', 4, 'AAC')],
                        'text':  [('delete', 0, 2), ('insert', 4, 'AAB')],
                        'op': ''}}
        da = diff_applier.DiffApplier(diff, None)
        node = {
            "label_id": "204",
            "title": "abcd",
            'node_type': APPENDIX
        }
        builder.diff_applier = da
        builder.process_node_title(node)
        self.assertEqual('<del>ab</del>cd<ins>AAC</ins>', node['header'])

    def test_node_title_no_diff(self):
        builder = HTMLBuilder(None, None, None)
        node = {
            "label_id": "204",
            "title": "abcd",
            'node_type': APPENDIX
        }
        builder.process_node_title(node)
        self.assertTrue('header' in node)
        self.assertEqual(node['title'], 'abcd')

    def test_generate_html(self):
        exex = Mock(spec=[])        # no attributes
        p_applier = Mock()
        p_applier.layers = {'exex': exex}
        p_applier.apply_layers.side_effect = lambda n: n    # identity
        builder = HTMLBuilder(None, p_applier, None)
        builder.tree = {'label': ['1234'], 'children': [],
                        'node_type': 'regtext', 'text': ''}
        builder.generate_html()
        #   No explosion so far means this works for most layers

        exex = Mock()               # "includes" any attribute
        p_applier = Mock()
        p_applier.layers = {'exex': exex}
        p_applier.apply_layers.side_effect = lambda n: n    # identity
        builder = HTMLBuilder(None, p_applier, None)
        builder.tree = {'label': ['1234'], 'children': [],
                        'node_type': 'regtext', 'text': ''}
        builder.generate_html()
        self.assertTrue(exex.preprocess_root.called)
        self.assertEqual(exex.preprocess_root.call_args[0][0],
                         builder.tree)

    def test_is_collapsed(self):
        for label, text in ((['111', '22', 'a'], '(a) '),
                            (['111', '22', 'xx'], ' (xx) '),
                            (['111', '22', 'a', '5'], '(5)')):
            node = {'label': label, 'text': text}
            self.assertTrue(HTMLBuilder.is_collapsed(node))

        for label, text in ((['111', '22', 'a'], '(b) '),
                            (['111', '22', ''], '(a) Some text'),
                            (['111', '22', 'a'], '  ')):
            node = {'label': label, 'text': text}
            self.assertFalse(HTMLBuilder.is_collapsed(node))

    def test_human_label(self):
        self.assertEqual(
            '111', HTMLBuilder.human_label({'label': ['111']}))
        self.assertEqual(
            '111-22-33-aa',
            HTMLBuilder.human_label({'label': ['111', '22', '33', 'aa']}))


class CFRHTMLBuilderTest(TestCase):
    def test_list_level_interpretations(self):
        builder = CFRHTMLBuilder(None, None, None)

        parts = ['101', '12', 'a', 'Interp', '1']
        node_type = INTERP

        result = builder.list_level(parts, node_type)
        self.assertEquals(result, 1)

        parts.append('j')
        result = builder.list_level(parts, node_type)
        self.assertEquals(result, 2)

        parts.append('B')
        result = builder.list_level(parts, node_type)
        self.assertEquals(result, 3)

    def test_list_level_appendices(self):
        builder = CFRHTMLBuilder(None, None, None)

        parts = ['101', 'A', '1', 'a']
        node_type = APPENDIX

        result = builder.list_level(parts, node_type)
        self.assertEquals(result, 1)

        parts.append('2')
        result = builder.list_level(parts, node_type)
        self.assertEquals(result, 2)

        parts.append('k')
        result = builder.list_level(parts, node_type)
        self.assertEquals(result, 3)

        parts.append('B')
        result = builder.list_level(parts, node_type)
        self.assertEquals(result, 4)

    def test_list_level_regulations(self):
        builder = CFRHTMLBuilder(None, None, None)

        parts = ['101', '1', 'a']
        node_type = REGTEXT

        result = builder.list_level(parts, node_type)
        self.assertEquals(result, 1)

        parts.append('2')
        result = builder.list_level(parts, node_type)
        self.assertEquals(result, 2)

        parts.append('k')
        result = builder.list_level(parts, node_type)
        self.assertEquals(result, 3)

        parts.append('B')
        result = builder.list_level(parts, node_type)
        self.assertEquals(result, 4)

    def test_list_level_regulations_no_level(self):
        builder = CFRHTMLBuilder(None, None, None)

        parts = ['101', '1']
        node_type = REGTEXT

        result = builder.list_level(parts, node_type)
        self.assertEquals(result, 0)

    def test_interp_node_with_citations(self):
        inline, p, sr = Mock(), Mock(), Mock()
        builder = CFRHTMLBuilder(inline, p, sr)

        node = {
            'text': 'Interpretation with a link',
            'children': [],
            'node_type': INTERP,
            'label': ['999', '5', 'Interp']
        }
        p.apply_layers.return_value = node
        inline.get_layer_pairs.return_value = []
        sr.get_layer_pairs.return_value = []
        builder.process_node(node)
        layer_parameters = inline.get_layer_pairs.call_args[0]
        self.assertEqual('Interpretation with a link', layer_parameters[1])
        self.assertEqual('999-5-Interp', layer_parameters[0])

    def test_no_section_sign(self):
        text = CFRHTMLBuilder.section_space(' abc')
        self.assertEquals(text, ' abc')
        self.assertTrue(True)

    def test_modify_interp_node(self):
        node = {
            'node_type': INTERP,
            'label': ['872', '22', 'Interp'],
            'children': [{'label': ['872', '22', 'Interp', '1']},
                         {'label': ['872', '22', 'a', 'Interp']},
                         {'label': ['872', '22', 'b', 'Interp']}]
        }
        builder = CFRHTMLBuilder(None, None, None)
        builder.modify_interp_node(node)
        self.assertTrue(node['section_header'])
        self.assertEqual(node['header_children'],
                         [{'label': ['872', '22', 'a', 'Interp']},
                          {'label': ['872', '22', 'b', 'Interp']}])
        self.assertEqual(node['par_children'],
                         [{'label': ['872', '22', 'Interp', '1']}])

        node['label'] = ['872', '222', 'a', 'Interp']
        builder.modify_interp_node(node)
        self.assertFalse(node['section_header'])

    def test_modify_interp_node_header(self):
        node = {
            'children': [],
            'header': 'This interprets 22(a), a paragraph',
            'label': ['872', '22', 'a', 'Interp'],
            'node_type': INTERP,
        }
        icl = InternalCitationLayer(None)
        icl.sectional = True
        ila = InlineLayersApplier()
        ila.add_layer(icl)
        builder = CFRHTMLBuilder(ila, None, None)

        builder.modify_interp_node(node)
        self.assertEqual('This interprets '
                         + icl.render_url(['872', '22', 'a'], '22(a)')
                         + ', a paragraph', node['header_markup'])

        node['label'] = ['872', '22']
        builder.modify_interp_node(node)
        self.assertEqual(node['header'], node['header_markup'])

    def test_process_node_title_section_space_diff(self):
        """" Diffs and sections spaces need to place nicely together. """
        builder = CFRHTMLBuilder(None, None, None)
        diff = {'204': {'title': [('delete', 7, 9), ('insert', 10, 'AAC')],
                        'text':  [('delete', 0, 2), ('insert', 4, 'AAB')],
                        'op': ''}}
        da = diff_applier.DiffApplier(diff, None)
        node = {
            "label_id": u"204",
            "title": u"§ 101.6 abcd",
            'node_type': APPENDIX
        }
        builder.diff_applier = da
        builder.process_node_title(node)
        self.assertEqual(
            u'§&nbsp;101.6<del> a</del>b<ins>AAC</ins>cd', node['header'])

    def test_human_label(self):
        self.assertEqual(
            'Regulation 111', CFRHTMLBuilder.human_label({'label': ['111']}))
        self.assertEqual(
            u'§ 111.22(f)',
            CFRHTMLBuilder.human_label({'label': ['111', '22', 'f']}))


class PreambleHTMLBuilderTest(TestCase):
    def setUp(self):
        inline, par, sr = Mock(), Mock(), Mock()
        inline.get_layer_pairs.return_value = []
        par.apply_layers.side_effect = lambda x: x
        sr.get_layer_pairs.return_value = []

        self.builder = PreambleHTMLBuilder(inline, par, sr)

    def test_human_label(self):
        self.assertEqual(
            'FR #111_22',
            PreambleHTMLBuilder.human_label({'label': ['111_22']}))
        self.assertEqual(
            'Section III.A.ii.4',
            PreambleHTMLBuilder.human_label({
                'label': ['111_22', 'III', 'A', 'ii', '4'],
                'indexes': [2, 0, 1, 3],
            }),
        )
        self.assertEqual(
            'Section III.A.ii.4 Paragraph 3.5',
            PreambleHTMLBuilder.human_label({
                'label': ['111_22', 'III', 'A', 'ii', '4', 'p3', 'p7'],
                'indexes': [2, 0, 1, 3, 2, 4]
            }),
        )

    def test_accepts_comment(self):
        """All of the preamble can be commented on. Some of it is called
        out"""
        node = {'label': ['ABCD_123', 'II', 'B', 'p4'], 'text': 'Something',
                'node_type': 'preamble', 'children': []}
        self.builder.process_node(node)
        self.assertTrue(node.get('accepts_comments'))
        self.assertFalse(node.get('comments_calledout'))

        node = {'title': 'B. Has a title', 'label': ['ABCD_123', 'II', 'B'],
                'text': 'Something', 'node_type': 'preamble', 'children': []}
        self.builder.process_node(node)
        self.assertTrue(node.get('accepts_comments'))
        self.assertTrue(node.get('comments_calledout'))


class CFRChangeHTMLBuilderTests(TestCase):
    def setUp(self):
        inline, par, sr = Mock(), Mock(), Mock()
        inline.get_layer_pairs.return_value = []
        par.apply_layers.side_effect = lambda x: x
        sr.get_layer_pairs.return_value = []
        diffs = DiffApplier({'111-22-a': {'op': 'deleted'}}, '111-22')
        self.builder = CFRChangeHTMLBuilder(inline, par, sr, diffs)

    def test_accepts_comment(self):
        """We can only comment on changed paragraphs"""
        node = {'label': ['111', '21', 'a'], 'text': 'Something',
                'node_type': 'regtext', 'children': []}
        self.builder.process_node(node)
        self.assertFalse(node.get('accepts_comments'))
        self.assertEqual(node.get('stars_collapse'), 'full')

        node['label'] = ['111', '22']
        self.builder.process_node(node)
        self.assertFalse(node.get('accepts_comments'))
        self.assertEqual(node.get('stars_collapse'), 'inline')

        node['label'] = ['111', '22', 'a', '5']
        self.builder.process_node(node)
        self.assertFalse(node.get('accepts_comments'))
        self.assertEqual(node.get('stars_collapse'), 'full')

        node['label'] = ['111', '22', 'a']
        self.builder.process_node(node)
        self.assertTrue(node.get('accepts_comments'))
        self.assertEqual(node.get('stars_collapse'), 'none')

    def test_preprocess(self):
        diffs = DiffApplier({'111-22-a': {'op': 'deleted'},
                             '111-33-b-5-v': {'op': 'deleted'}}, '111-22')
        builder = CFRChangeHTMLBuilder(
            self.builder.inline_applier, self.builder.p_applier,
            self.builder.search_applier, diffs)
        self.assertEqual(
            builder.diff_paths,
            {('111',), ('111', '22'), ('111', '22', 'a'), ('111', '33'),
             ('111', '33', 'b'), ('111', '33', 'b', '5'),
             ('111', '33', 'b', '5', 'v')})



from itertools import takewhile

from django.core.urlresolvers import reverse, NoReverseMatch

from regulations.generator.node_types import to_markup_id
from regulations.generator.toc import fetch_toc


class SectionUrl(object):
    """With few exceptions, users are expected to browse the regulation by
    traversing regtext sections, appendices, and subterps (split
    interpretations). This object will deduce, from a version and citation,
    to which section/appendix/subterp to link, a task greatly complicated by
    subterps. Importantly, this object keeps a cache of looked up info;
    reusing an instance is significantly faster than using static methods."""
    def __init__(self):
        self.rev_cache = {}
        self.toc_cache = {}

    def view_label_id(self, citation, version):
        # Subterps, collections of interps of whole subparts, etc.
        if 'Interp' in citation and ('Subpart' in citation
                                     or 'Appendices' in citation):
            label = '-'.join(citation)
        elif 'Interp' in citation:
            label = self.interp(citation, version)
        else:
            label = '-'.join(citation[:2])
        return label

    def fetch(self, citation, version, sectional):
        key = (tuple(citation), version, sectional)
        if key not in self.rev_cache:
            url = ''

            if sectional:
                view_name = 'chrome_section_view'
                if len(citation) > 1 and citation[1] == 'Interp':
                    view_name = 'chrome_paragraph_view'
                elif 'Interp' in citation:
                    view_name = 'chrome_subterp_view'
                label = self.view_label_id(citation, version)

                try:
                    url = reverse(view_name, args=(label, version))
                except NoReverseMatch:
                    # XXX We have some errors in our layers. Once those are
                    # fixed, we need to revisit this.
                    pass
            self.rev_cache[key] = url + '#' + '-'.join(to_markup_id(citation))
        return self.rev_cache[key]

    def interp(self, citation, version):
        """Subterps throw a big monkey-wrench into things. Citations to
        interpretations must be converted into their corresponding subterp,
        which requires loading the toc."""
        reg_part = citation[0]
        key = (reg_part, version)
        prefix = list(takewhile(lambda l: l != 'Interp', citation))
        prefix.append('Interp')
        if key not in self.toc_cache:
            self.toc_cache[key] = fetch_toc(reg_part, version)

        prefix_section = prefix[:2]
        for el in self.toc_cache[key]:
            if el['index'] == prefix_section and el.get('is_section'):
                # No subpart
                return reg_part + '-Subpart-Interp'
            for sub in el.get('sub_toc', []):
                #   In a subpart
                if sub['index'] == prefix_section and el.get('is_subpart'):
                    return '-'.join(el['index'][:3] + ['Interp'])
                #   In a subterp
                if sub['index'] == prefix:
                    return '-'.join(prefix)
                #   Other match: interpretation header
                if sub['index'] == citation[:len(sub['index'])]:
                    return sub['section_id']
        # Couldn't find it; most likely in the appendix
        return reg_part + '-Appendices-Interp'

    @staticmethod
    def of(citation, version, sectional):
        return SectionUrl().fetch(citation, version, sectional)



from unittest import TestCase
from mock import patch

from django.test import RequestFactory

from regulations.views import error_handling


class ErrorHandlingTest(TestCase):

    @patch('regulations.views.error_handling.api_reader')
    def test_check_regulation_none(self, api_reader):
        api_reader.ApiReader.return_value.regversions.return_value = None

        self.assertRaises(
            error_handling.MissingContentException,
            error_handling.check_regulation, '204')

    @patch('regulations.views.error_handling.api_reader')
    def test_check_regulation_exists(self, api_reader):
        api_reader.ApiReader.return_value.regversions.return_value = [1]

        result = error_handling.check_regulation('204')
        self.assertEqual(result, None)

    @patch('regulations.views.error_handling.api_reader')
    def test_check_version(self, api_reader):
        api_reader.ApiReader.return_value.regversions.return_value =\
            {'versions': [{'version': '2'}]}
        result = error_handling.check_version('204', '2')
        self.assertTrue(result)

    @patch('regulations.views.error_handling.api_reader')
    def test_check_no_version(self, api_reader):
        api_reader.ApiReader.return_value.regversions.return_value =\
            {'versions': [{'version': '3'}]}
        result = error_handling.check_version('204', '2')
        self.assertFalse(result)

    def test_handle_generic_404(self):
        request = RequestFactory().get('/fake-path')
        result = error_handling.handle_generic_404(request)
        self.assertEqual(result.status_code, 404)
        self.assertTrue(b'Regulation content not found' in result.content)

    @patch('regulations.views.error_handling.add_to_chrome')
    @patch('regulations.views.error_handling.api_reader')
    def test_handle_missing_section_404(self, api_reader, add_to_chrome):
        api_reader.ApiReader.return_value.regversions.return_value =\
            {'versions': [{'version': '2', 'by_date': '2013-03-26'}]}
        add_to_chrome.return_value = None

        request = RequestFactory().get('/fake-path')

        extra_content = {'passed': 1, 'env': 'source'}
        response = error_handling.handle_missing_section_404(
            request, '204-1', '2', extra_content)
        self.assertEqual(response, None)
        self.assertTrue(add_to_chrome.called)



import copy
from collections import Counter, deque, namedtuple

from regulations.generator.layers import tree_builder


class DiffApplier(object):
    """ Diffs between two versions of a regulation are represented in our
    particular JSON format. This class applies that diff to the older version
    of the regulation, generating HTML that clearly shows the changes between
    old and new. """

    INSERT = u'insert'
    DELETE = u'delete'
    EQUAL = u'equal'

    DELETED_OP = 'deleted'
    ADDED_OP = 'added'
    MODIFIED_OP = 'modified'

    def __init__(self, diff_json, label_requested):
        self.diff = diff_json
        # label_requested is the regulation label for which a diff is
        # requested.
        self.label_requested = label_requested

    def deconstruct_text(self, original):
        self.oq = [deque([c]) for c in original]

    def insert_text(self, pos, new_text):
        if pos == len(self.oq):
            self.oq[pos-1].extend(['<ins>', new_text, '</ins>'])
        else:
            self.oq[pos].appendleft('<ins>' + new_text + '</ins>')

    def delete_text(self, start, end):
        self.oq[start].appendleft('<del>')
        self.oq[end-1].append('</del>')

    def get_text(self):
        return ''.join([''.join(d) for d in self.oq])

    def delete_all(self, text):
        """ Mark all the text passed in as deleted. """
        return '<del>' + text + '</del>'

    def add_all(self, text):
        """ Mark all the text passed in as deleted. """
        return '<ins>' + text + '</ins>'

    _LabelOp = namedtuple('LabelOp', ['label', 'op'])

    def set_child_labels(self, node):
        """As we display removed, added, and unchanged nodes, the children of
        a node will contain all three types. Pull the 'child_ops' data to
        derive the correct order of these combined children"""
        instructs = self.diff.get('-'.join(node['label']), {})
        if 'child_labels' not in instructs and 'child_ops' in instructs:
            original_labels = ['-'.join(c['label']) for c in node['children']]
            label_ops = []
            for op, start, end_or_labels in instructs['child_ops']:
                if isinstance(end_or_labels, list):
                    labels = end_or_labels
                else:
                    labels = original_labels[start:end_or_labels]
                label_ops.extend(DiffApplier._LabelOp(l, op) for l in labels)
            label_ops = self.remove_moved_labels(label_ops)
            node['child_labels'] = [lo.label for lo in label_ops]

    @classmethod
    def has_moved(cls, label_op, seen_count):
        """A label is moved if it's been deleted in one position but added int
        another"""
        if seen_count[label_op.label] == 1:
            return False
        else:   # We want to keep the final destination (INSERT)
            return label_op.op != cls.INSERT

    def remove_moved_labels(self, label_ops):
        """If a label has been moved, we will display it in the new position"""
        seen_count = Counter(label_op.label for label_op in label_ops)

        return [lo for lo in label_ops if not self.has_moved(lo, seen_count)]

    def add_nodes_to_tree(self, original, adds):
        """ Add all the nodes from new_nodes into the original tree. """
        tree_hash = tree_builder.build_tree_hash(original)
        for node in tree_hash.values():
            self.set_child_labels(node)

        for label, node in adds.queue:
            p_label = '-'.join(tree_builder.parent_label(node))
            if tree_builder.parent_in_tree(p_label, tree_hash):
                tree_builder.add_node_to_tree(node, p_label, tree_hash)
                adds.delete(label)
            else:
                parent = adds.find(p_label)
                if parent:
                    tree_builder.add_child(parent[1], node)
                else:
                    original.update(node)

    def is_child_of_requested(self, label):
        """ Return true if the label is a child of the requested label.  """
        req = self.label_requested
        if 'Interp' in label and 'Interp' in req:
            # Sub-paragraph
            if label.startswith(req + '-'):
                return True
            # The parent must not be a sub paragraph if the prefixes differ
            if not req.endswith('-Interp'):
                return False
            req_interpreting = req[:req.find('-Interp')]
            child_interpreting = label[:label.find('-Interp')]
            return child_interpreting.startswith(req_interpreting + '-')
        elif 'Interp' not in label and 'Interp' not in req:
            return label.startswith(req + '-')
        return False

    def relevant_added(self, label):
        """ Get the operations that add nodes, for the requested
        section/pargraph. """

        if (self.diff[label]['op'] == self.ADDED_OP
            and (label == self.label_requested
                 or self.is_child_of_requested(label))):
            return True

    def tree_changes(self, original_tree):
        """ Apply additions to the regulation tree. """

        def node(diff_node, label):
            """ Take diff's specification of a node, and actually turn it into
            a regulation node. """

            node = copy.deepcopy(diff_node)
            node['children'] = []
            if 'title' in node and node['title'] is None:
                del node['title']
            return node

        new_nodes = [(label, node(self.diff[label]['node'], label))
                     for label in self.diff if self.relevant_added(label)]

        adds = tree_builder.AddQueue()
        adds.insert_all(new_nodes)

        self.add_nodes_to_tree(original_tree, adds)

    def apply_diff_changes(self, original, diff_list):
        """Account for modified text"""
        self.deconstruct_text(original)
        for d in diff_list:
            if d[0] == self.INSERT:
                _, pos, new_text = d
                self.insert_text(pos, new_text)
            if d[0] == self.DELETE:
                _, s, e = d
                self.delete_text(s, e)
            if isinstance(d[0], list):
                if d[0][0] == self.DELETE and d[1][0] == self.INSERT:
                    # Text replace scenario.
                    _, s, e = d[0]
                    self.delete_text(s, e)

                    _, _, new_text = d[1]

                    # Place the new text at the end of the delete for
                    # readability.
                    self.insert_text(e, new_text)
        return self.get_text()

    def apply_diff(self, original, label, component='text'):
        """Here we delete or add whole nodes in addition to passing to
        `apply_diff_changes` when text has been modified"""
        if label in self.diff:
            if self.diff[label]['op'] == self.DELETED_OP:
                return self.delete_all(original)
            if self.diff[label]['op'] == self.ADDED_OP:
                return self.add_all(original)
            if component in self.diff[label]:
                text_diffs = self.diff[label][component]
                return self.apply_diff_changes(original, text_diffs)
        return original



from django.core.management.base import BaseCommand
from eregs_cache import EregsCache
import sys


class Command(BaseCommand):
    args = ("<url where eregulations is located> <optional- comma seperated "
            "list of regulation locations to be parsed >")
    help = 'call every page in eregulations allowing the pages to be cached'

    def handle(self, *args, **options):

        regulations_arg = None
        if len(sys.argv) > 3:
            regulations_arg = sys.argv[3]

        EregsCache(sys.argv[2], regulations_arg)



from datetime import datetime
import re

import six
from django.template import Context


def convert_to_python(data):
    """Convert raw data (e.g. from json conversion) into the appropriate
    Python objects"""
    if isinstance(data, six.string_types):
        #   Dates
        if re.match(r'^\d{4}-\d{2}-\d{2}$', data):
            return datetime.strptime(data, '%Y-%m-%d')
    if isinstance(data, dict):
        new_data = {}
        for key in data:
            new_data[key] = convert_to_python(data[key])
        return new_data
    if isinstance(data, tuple):
        return tuple(map(convert_to_python, data))
    if isinstance(data, list):
        return list(map(convert_to_python, data))

    return data


def render_template(template, context):
    c = Context(context)
    return template.render(c).strip('\n')


def is_contained_in(child, parent):
    '''
        Return True if child is a child node of the parent.

        Node labels are hierarchical paths, with segments separated
        by '-'. As an edge case, a node label is also a child of itself.
    '''
    child_segments = child.split('-')
    parent_segments = parent.split('-')
    return child_segments[:len(parent_segments)] == parent_segments



import re

# <a> followed by another <a> without any intervening </a>s
# outer_link - partial outer element up to the inner link
# inner_content - content of the inner_link
link_inside_link_regex = re.compile(
    r"(?P<outer_link><a ((?!</a>).)*)<a .*?>(?P<inner_content>.*?)</a>",
    re.IGNORECASE | re.DOTALL)


def flatten_links(text):
    """
    Fix <a> elements that have embedded <a> elements by
    replacing the internal <a> element with its content.
    """

    while True:
        text, sub_count = link_inside_link_regex.subn(
            r"\g<outer_link>\g<inner_content>", text)
        if sub_count == 0:
            return text



from django.http import HttpResponse

from regulations.generator import api_reader
from regulations.generator.node_types import label_to_text
from regulations.views.chrome import ChromeView
from regulations.views.partial_sxs import ParagraphSXSView


class ChromeBreakawayView(ChromeView):
    """ Base class for views which wish to include chrome. """
    template_name = 'regulations/breakaway-chrome.html'

    def content(self, context):
        """Filled in by subclasses; content for the breakaway"""
        raise NotImplementedError

    def get_context_data(self, **kwargs):
        #   Skip ChromeView's implementation
        context = super(ChromeView, self).get_context_data(**kwargs)

        context['reg_part'] = context['label_id'].split('-')[0]
        context['version'] = self.request.GET.get('from_version')
        meta = api_reader.ApiReader().layer(
            'meta', 'cfr', context['reg_part'], context['version'])
        context['meta'] = meta[context['reg_part']][0]
        context['formatted_id'] = label_to_text(context['label_id'].split('-'))

        content = self.content(context)
        if isinstance(content, HttpResponse):  # error occurred
            return content
        context['partial_content'] = content

        return context


class ChromeSXSView(ChromeBreakawayView):
    """SXS content"""
    def content(self, context):
        partial_view = ParagraphSXSView.as_view()
        response = partial_view(self.request, label_id=context['label_id'],
                                notice_id=context['notice_id'])
        self._assert_good(response)
        response.render()
        return response.content



from django import http
from django.template import RequestContext, loader

from regulations.generator import api_reader
from regulations.generator.sidebar.help import Help as HelpSidebar
from regulations.generator.layers.utils import convert_to_python
from regulations.views.sidebar import SideBarView


class MissingContentException(Exception):
    """ This is essentially a generic 404. """
    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "MissingContentException"


class MissingSectionException(Exception):
    """" This is for when we suspect that we have the version requested, but
    maybe just not the label_id. """

    def __init__(self, label_id, version, context):
        self.label_id = label_id
        self.version = version
        self.context = context

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "MissingSectionException(%s, %s)" % (
            self.label_id, self.version)


def handle_generic_404(request):
    template = loader.get_template('regulations/generic_404.html')
    context = {'request_path': request.path}
    body = template.render(RequestContext(
        request, context))
    return http.HttpResponseNotFound(body, content_type='text/html')


def check_regulation(reg_part):
    """ If versions of the reg_part given don't exist, raise
    a MissingContentException(). """

    client = api_reader.ApiReader()
    vr = client.regversions(reg_part)

    if not vr:
        raise MissingContentException()


def check_version(label_id, version):
    """ We check if the version of this regulation exists, and the user is only
    referencing a section that does not exist. """

    reg_part = label_id.split('-')[0]
    client = api_reader.ApiReader()
    vr = client.regversions(reg_part)

    requested_version = [v for v in vr['versions'] if v['version'] == version]

    if len(requested_version) > 0:
        requested_version = convert_to_python(requested_version)
        return requested_version[0]


def add_to_chrome(body, context, request):
    chrome_template = loader.get_template('regulations/chrome.html')

    context['main_content'] = body
    sidebar_view = SideBarView.as_view(components=[HelpSidebar])
    sidebar_response = sidebar_view(request, label_id=context['label_id'],
                                    version=context['version'])
    context['sidebar_content'] = sidebar_response.render().content
    chrome_body = chrome_template.render(RequestContext(
        request, context))

    return http.HttpResponseNotFound(chrome_body, content_type='text/html')


def handle_missing_section_404(
        request, label_id, version, extra_context=None):

    req_version = check_version(label_id, version)
    if not req_version:
        return handle_generic_404(request)

    reg_section = label_id.split('-')[1]

    context = {
        'request_path': request.path,
        'reg_section': reg_section,
        'effective_date': req_version['by_date']
    }
    context.update(extra_context)

    template = loader.get_template('regulations/missing_section_404.html')
    body = template.render(RequestContext(
        request, context))

    return add_to_chrome(body, context, request)



import string
from django.template import loader

from regulations.generator.layers import utils
from regulations.generator.layers.base import SearchReplaceLayer


class KeyTermsLayer(SearchReplaceLayer):
    shorthand = 'keyterms'
    data_source = 'keyterms'
    _text_field = 'key_term'

    def __init__(self, layer):
        self.layer = layer
        self.template = loader.get_template('regulations/layers/key_term.html')

    def remove_punctuation(self, punctuated):
        translate_table = dict((ord(c), None) for c in string.punctuation)
        return punctuated.translate(translate_table)

    def replacements_for(self, original, data):
        key_term = {'key_term': original,
                    'phrase': self.remove_punctuation(original)}
        context = {'key_term': key_term}
        yield utils.render_template(self.template, context)



from django.core.cache import caches
from regulations.generator import api_client


_cache_key = '-'.join


class ApiReader(object):
    """ Access the regulations API. Either hit the cache, or if there's a miss,
    hit the API instead and cache the results. """

    def __init__(self):
        self.cache = caches['api_cache']
        self.client = api_client.ApiClient()

    def all_regulations_versions(self):
        """ Get all versions, for all regulations. """
        return self._get(['all_regulations_versions'], 'regulation')

    def regversions(self, label):
        return self._get(
            ['regversions', label],
            'regulation/%s' % label)

    def cache_root_and_interps(self, reg_tree, version, is_root=True):
        """We will re-use the root tree at multiple points during page
        rendering, so cache it now. If caching an interpretation, also store
        child interpretations with titles (so that, when rendering slide-down
        interpretations, we don't perform additional fetches)"""
        if is_root or reg_tree.get('title'):
            tree_id = '-'.join(reg_tree['label'])
            cache_key = _cache_key(['regulation', tree_id, version])
            self.cache.set(cache_key, reg_tree)

        for child in reg_tree['children']:
            if child.get('node_type') == 'interp':
                self.cache_root_and_interps(child, version, False)

    def regulation(self, label, version):
        cache_key = _cache_key(['regulation', label, version])
        cached = self.cache.get(cache_key)

        if cached is not None:
            return cached
        else:
            regulation = self.client.get('regulation/%s/%s' % (label, version))
            # Add the tree to the cache
            if regulation:
                self.cache_root_and_interps(regulation, version)
                return regulation

    def _get(self, cache_key_elements, api_suffix, api_params={}):
        """ Retrieve from the cache whenever possible, or get from the API """
        cache_key = _cache_key(cache_key_elements)
        cached = self.cache.get(cache_key)

        if cached is not None:
            return cached
        else:
            element = self.client.get(api_suffix, api_params)
            self.cache.set(cache_key, element)
            return element

    def layer(self, layer_name, doc_type, label_id, version=None):
        """When retrieving layer data, we cheat a bit -- we always retrieve
        layer data corresponding to the "root" of the document, rather than
        only a subnode. We also must convert to the API format, where any
        version information is prefixed to doc_id"""
        root = label_id.split('-')[0]
        if version is None:
            doc_id = root
        else:
            doc_id = '{}/{}'.format(version, root)
        result = self._get(
            ('layer', layer_name, doc_type, root, str(version)),
            'layer/{}/{}/{}'.format(layer_name, doc_type, doc_id))
        # To remove - also try the old format for CFR layers; the API may not
        # have been updated
        if result is None and doc_type == 'cfr':
            result = self._get(
                ('layer', layer_name, doc_type, root, str(version)),
                'layer/{}/{}/{}'.format(layer_name, root, version))
        return result

    def diff(self, label, older, newer):
        """ End point for diffs. """
        return self._get(
            ['diff', label, older, newer],
            "diff/%s/%s/%s" % (label, older, newer))

    def notices(self, part=None):
        """ End point for notice searching. Right now just a list. """
        if part:
            return self._get(
                ['notices', part],
                'notice',
                {'part': part})
        else:
            return self._get(
                ['notices'],
                'notice')

    def notice(self, fr_document_number):
        """ End point for retrieving a single notice. """
        return self._get(
            ['notice', fr_document_number],
            'notice/%s' % fr_document_number)

    def search(self, query, doc_type='cfr', version=None, regulation=None,
               **kwargs):
        """Search via the API. Never cache these (that's the duty of the search
        index)"""
        params = {'q': query}
        if version:
            params['version'] = version
        if regulation:
            params['regulation'] = regulation
        params.update(kwargs)
        return self.client.get('/'.join(['search', doc_type]), params)

    def preamble(self, doc_number):
        return self._get(
            ['preamble', doc_number], 'preamble/{}'.format(doc_number))



from regulations.generator.layers.base import LayerBase
from regulations.generator.layers.utils import convert_to_python


class MetaLayer(LayerBase):
    shorthand = 'meta'
    data_source = 'meta'
    layer_type = LayerBase.PARAGRAPH

    def __init__(self, layer_data):
        self.layer_data = convert_to_python(layer_data)

    def apply_layer(self, text_index):
        """Return a pair of field-name (meta) + the layer data"""
        if self.layer_data.get(text_index):
            return 'meta', self.layer_data[text_index][0]



"""Particularly when displaying bullets, it can be immensely helpful to
construct a list _within_ the template. Django's templates don't provide this
out of the box, so we need a new template tag. Suggestion from:
    http://stackoverflow.com/a/34407158

Usage:
    {% to_list "a" "b" "c" as my_list %}
    {% for el in my_list %}
        {{ el }}
    {% endfor %}
"""
from django import template

register = template.Library()


@register.assignment_tag
def to_list(*args):
    return args



import codecs
import os
import shutil
import subprocess   # nosec - see usage below

from django.conf import settings
from django.core.management.base import BaseCommand
from django.contrib.staticfiles.finders import get_finders
from django.contrib.staticfiles.storage import StaticFilesStorage

import regulations


"""
This command compiles the frontend for regulations-site after using the Django
``collectstatic`` command to override specific files, and places the output in
a directory named ``compiled`` at the root level of the project.

For example, the atf-eregs project uses ``regulations-site`` as a library and
overrides the contents of the
``regulations/static/regulations/css/less/mixins.less`` file. The
``atf_eregs/regulations/static/regulations/css/less/mixins.less`` file will be
copied over the base file from ``regulations-site`` and put into a build
directory by ``collectstatic``, and then the frontend build commands are run,
and the results are copied into ``compiled``, which can then be used as the
static directory for the CSS, font, JavaScript, and image assets.

"""


class Command(BaseCommand):
    help = 'Build the frontend, including local overrides.'
    BUILD_DIR = "./frontend_build"
    TARGET_DIR = "./compiled/regulations"

    def add_arguments(self, parser):
        parser.add_argument('--no-install', dest='install',
                            action='store_false')
        parser.set_defaults(install=True)

    def find_regulations_directory(self):
        child = regulations.__file__
        child_dir = os.path.split(child)[0]
        return os.path.split(child_dir)[0] or "."   # if regulations is local

    def remove_dirs(self):
        """Remove existing output dirs"""
        if os.path.exists(self.TARGET_DIR):
            shutil.rmtree(self.TARGET_DIR)
        # Delete everything in BUILD_DIR except node_modules, which we use for
        # caching the downloaded node libraries
        if os.path.exists(self.BUILD_DIR):
            all_content = [os.path.join(self.BUILD_DIR, f)
                           for f in os.listdir(self.BUILD_DIR)]
            files = filter(os.path.isfile, all_content)
            dirs = filter(os.path.isdir, all_content)
            for f in files:
                os.remove(f)
            for d in dirs:
                if d != os.path.join(self.BUILD_DIR, 'node_modules'):
                    shutil.rmtree(d)
        else:
            os.mkdir(self.BUILD_DIR)

    def copy_configs(self):
        """Copy over configs from regulations"""
        regulations_directory = self.find_regulations_directory()
        frontend_files = (
            "package.json",
            "Gruntfile.js",
            ".eslintrc"
        )
        for f_file in frontend_files:
            source = "%s/%s" % (regulations_directory, f_file)
            shutil.copy(source, "%s/" % self.BUILD_DIR)
        with codecs.open("%s/config.json" % self.BUILD_DIR, "w",
                         encoding="utf-8") as f:
            f.write('{"frontEndPath": "static/regulations"}')

    def _input_files(self):
        """Fetch all of the static files from the installed apps. Yield them
        as pairs of (path, file)"""
        files_seen = set()
        pairs = (pr for finder in get_finders() for pr in finder.list([".*"]))
        for path, storage in pairs:
            # Prefix the relative path if the source storage contains it
            if getattr(storage, 'prefix', None):
                prefixed_path = os.path.join(storage.prefix, path)
            else:
                prefixed_path = path
            if prefixed_path in files_seen:
                self.stdout.write(
                    "Using override for {}\n".format(prefixed_path))
            else:
                files_seen.add(prefixed_path)
                with storage.open(path) as source_file:
                    yield (prefixed_path, source_file)

    def collect_files(self):
        """Find and write static files. Along the way ignore the "compiled"
        directory, if present"""
        write_storage = StaticFilesStorage(self.BUILD_DIR + "/static/")
        original_dirs = settings.STATICFILES_DIRS
        settings.STATICFILES_DIRS = [s for s in original_dirs
                                     if s != 'compiled']
        for prefixed_path, source_file in self._input_files():
            write_storage.save(prefixed_path, source_file)
        settings.STATICFILES_DIRS = original_dirs

    def build_frontend(self, install=True, dev=False):
        """Shell out to npm for building the frontend files"""
        cmd = 'build-dev' if dev else 'build-dist'
        # Safe because: we're not passing user input into these processes
        if install:
            subprocess.check_call(  # nosec
                ['npm', 'install'], cwd=self.BUILD_DIR)
        subprocess.check_call(['grunt', cmd], cwd=self.BUILD_DIR)   # nosec

    def cleanup(self):
        shutil.copytree("%s/static/regulations" % self.BUILD_DIR,
                        self.TARGET_DIR)

    def handle(self, **options):
        self.remove_dirs()
        self.copy_configs()
        self.collect_files()
        self.build_frontend(install=options['install'], dev=settings.JS_DEBUG)
        self.cleanup()



# vim: set encoding=utf-8
import itertools
import logging

from regulations.generator import api_reader, generator
from regulations.generator.layers.tree_builder import roman_nums
from regulations.generator.layers.utils import convert_to_python
from regulations.generator.toc import fetch_toc


logger = logging.getLogger(__name__)


def to_roman(number):
    """ Convert an integer to a roman numeral """
    romans = list(itertools.islice(roman_nums(), 0, number + 1))
    return romans[number - 1]


def get_layer_list(names):
    layer_names = generator.LayerCreator.LAYERS
    return set(l.lower() for l in names.split(',') if l.lower() in layer_names)


def regulation_meta(cfr_part, version):
    """ Return the contents of the meta layer, without using a tree. """
    layer_data = api_reader.ApiReader().layer('meta', 'cfr', cfr_part, version)
    if layer_data is None:
        logger.warning("404 when fetching Meta for %s@%s", cfr_part, version)
        layer_data = {}
    if not layer_data.get(cfr_part):
        logger.warning("Empty meta data for %s@%s. Misparsed?",
                       cfr_part, version)

    meta = {}
    for data in layer_data.get(cfr_part, []):
        meta.update(data)
    return convert_to_python(meta)


def layer_names(request):
    """Determine which layers are currently active by looking at the request"""
    if 'layers' in request.GET.keys():
        return get_layer_list(request.GET['layers'])
    else:
        return generator.LayerCreator.LAYERS.keys()


def first_section(reg_part, version):
    """ Use the table of contents for a regulation, to get the label of the
    first section of the regulation. In most regulations, this is -1, but in
    some it's -101. """

    toc = fetch_toc(reg_part, version, flatten=True)
    return toc[0]['section_id']


def make_sortable(string):
    """Split a string into components, converting digits into ints so sorting
    works as we would expect"""
    if not string:      # base case
        return tuple()
    elif string[0].isdigit():
        prefix = "".join(itertools.takewhile(lambda c: c.isdigit(), string))
        return (int(prefix),) + make_sortable(string[len(prefix):])
    else:
        prefix = "".join(itertools.takewhile(lambda c: not c.isdigit(),
                                             string))
        return (prefix,) + make_sortable(string[len(prefix):])



# vim: set encoding=utf-8
from django.http import Http404
from django.views.generic.base import TemplateView

from regulations.generator import generator
from regulations.generator.html_builder import CFRHTMLBuilder
from regulations.generator.node_types import EMPTYPART, REGTEXT, label_to_text
from regulations.views import navigation, utils


def generate_html(regulation_tree, layer_appliers):
    builder = CFRHTMLBuilder(*layer_appliers)
    builder.tree = regulation_tree
    builder.generate_html()
    return builder


class PartialView(TemplateView):
    """Base class of various partial markup views. sectional_links indicates
    whether this view should use section links (url to a path) or just hash
    links (to an anchor on the page)"""

    sectional_links = True

    def determine_appliers(self, label_id, version):
        """Figure out which layers to apply by checking the GET args"""
        layer_creator = generator.LayerCreator()
        layer_creator.add_layers(utils.layer_names(self.request), 'cfr',
                                 label_id, self.sectional_links, version)
        return layer_creator.get_appliers()

    def get_context_data(self, **kwargs):
        context = super(PartialView, self).get_context_data(**kwargs)

        label_id = context['label_id']
        version = context['version']

        tree = generator.get_tree_paragraph(label_id, version)
        if tree is None:
            raise Http404

        inline_applier, p_applier, s_applier = self.determine_appliers(
            label_id, version)

        builder = generate_html(tree, (inline_applier, p_applier, s_applier))
        return self.transform_context(context, builder)


class PartialSectionView(PartialView):
    """ Single section of reg text """
    template_name = 'regulations/regulation-content.html'

    def section_navigation(self, label, version):
        nav_sections = navigation.nav_sections(label, version)
        if nav_sections:
            p_sect, n_sect = nav_sections

            return {'previous': p_sect, 'next': n_sect,
                    'page_type': 'reg-section'}

    def transform_context(self, context, builder):
        child_of_root = builder.tree
        #   Add a layer to account for subpart if this is regtext
        if builder.tree['node_type'] == REGTEXT:
            child_of_root = {
                'node_type': EMPTYPART,
                'children': [builder.tree]}
        context['markup_page_type'] = 'reg-section'
        context['tree'] = {'children': [child_of_root]}
        context['navigation'] = self.section_navigation(
            context['label_id'], context['version'])

        return context


class PartialParagraphView(PartialSectionView):
    """ Single paragraph of a regtext """
    def transform_context(self, context, builder):
        node = builder.tree
        # Wrap with layers until we reach a section
        while len(node['label']) > 2:
            node = {'node_type': node['node_type'],
                    'children': [node],
                    'label': node['label'][:-1]}

        # added to give the proper parent container ID
        # when interp headers are rendered
        node['markup_id'] = context['label_id']

        # One more layer for regtext
        if node['node_type'] == REGTEXT:
            node = {'node_type': EMPTYPART,
                    'children': [node],
                    'label': node['label'][:1] + ['Subpart']}

        context['markup_page_type'] = 'reg-section'
        context['tree'] = {'children': [node], 'label': node['label'][:1],
                           'node_type': REGTEXT}
        context['navigation'] = self.section_navigation(
            context['label_id'], context['version'])
        return context


class PartialDefinitionView(PartialView):
    """ Single paragraph of a regtext formatted for display
        as an inline interpretation """

    template_name = "regulations/partial-definition.html"

    def transform_context(self, context, builder):
        context['node'] = builder.tree
        context['formatted_label'] = label_to_text(
            builder.tree['label'], True, True)
        context['node']['section_id'] = '%s-%s' % (
            builder.tree['label'][0], builder.tree['label'][1])
        return context


class PartialRegulationView(PartialView):
    """ Entire regulation without chrome """

    template_name = 'regulations/regulation-content.html'
    sectional_links = False

    def transform_context(self, context, builder):
        context['tree'] = builder.tree
        return context



# vim: set encoding=utf-8
import unittest

from selenium.webdriver.support.ui import WebDriverWait

from regulations.uitests.base_test import BaseTest
from regulations.uitests import utils


class DefinitionTest(BaseTest, unittest.TestCase):

    job_name = 'Definitions test'

    def test_definition(self):
        self.driver.set_window_size(1024, 600)
        self.driver.get(self.test_url + '/1005-1/2012-12121')
        html = self.driver.find_element_by_tag_name('html')
        WebDriverWait(self.driver, 30).until(
            lambda driver: 'selenium-start' in html.get_attribute('class'))
        definition_link = self.driver.find_element_by_xpath(
            '//*[@id="1005-1-a"]//a')
        # term link should have correct data attr
        self.assertIn(
            '1005-2-a-1', definition_link.get_attribute('data-definition'))

        definition_link.click()

        # term link should get active class
        self.assertIn('active', definition_link.get_attribute('class'))

        definition = self.driver.find_element_by_xpath('//*[@id="1005-2-a-1"]')
        definition_close_button = self.driver.find_element_by_xpath(
            '//*[@id="1005-2-a-1"]/div[1]/h4/a')

        # definition should appear in sidebar
        self.assertGreater(len(definition.text), 20)
        definition_term = self.driver.find_element_by_xpath(
            '//*[@id="1005-2-a-1"]/div[3]/p/dfn')
        self.assertIn(u'\u201cvoided tosser\u201d', definition_term.text)

        definition_close_button.click()
        # definition should close
        self.assertFalse('active' in definition_link.get_attribute('class'))

        definition_link.click()

        # continue link should load full def
        definition_cont_link = self.driver.find_element_by_xpath(
            '//*[@id="1005-2-a-1"]/div[3]/a[1]')
        definition_cont_link.click()
        WebDriverWait(self.driver, 30).until(
            lambda driver: driver.find_element_by_xpath('//*[@id="1005-2"]'))

        # test definition scope notifications
        toc_toggle = self.driver.find_element_by_css_selector('#panel-link')
        toc_toggle.click()

        toc_1005_3 = self.driver.find_element_by_xpath(
            '//*[@id="toc"]/ol/li[3]/a')
        toc_1005_3.click()

        # close toc
        toc_toggle.click()

        # load 1005.3, open definition
        new_definition_link = self.driver.find_element_by_xpath(
            '//*[@id="1005-3-a"]//a[1]')
        self.driver.execute_script('window.scrollTo(0, 0);')
        new_definition_link.click()
        self.driver.find_element_by_xpath('//*[@id="1005-2-b-1"]')

        # navigate back to 1005.1
        toc_toggle.click()
        self.driver.find_element_by_xpath('//*[@id="toc"]/ol/li[1]/a').click()

        # make sure that the scope notice displays
        self.driver.find_element_by_xpath('//*[@id="1005-2-b-1"]/div[2]/div')

        # go to 1005-1-a
        toc_toggle.click()
        utils.scroll_to(self.driver, '#1005-1-a')

        definition_update_link = self.driver.find_element_by_css_selector(
            '.update-definition')
        definition_text = self.driver.find_element_by_css_selector(
            '.definition-text')

        # make sure text is grayed out
        self.assertIn('inactive', definition_text.get_attribute('class'))

        # load in scope definition
        definition_update_link.click()
        WebDriverWait(self.driver, 10).until(
            lambda driver: driver.find_element_by_xpath(
                '//*[@id="1005-2-a-1"]'))



from django.template import loader, Context

from regulations.generator.layers.base import SearchReplaceLayer


class FormattingLayer(SearchReplaceLayer):
    shorthand = 'formatting'
    data_source = 'formatting'
    DATA_TYPES = ('table', 'fence', 'subscript', 'dash', 'footnote',
                  'superscript')

    def __init__(self, layer):
        self.layer = layer
        self.tpls = {
            key: loader.get_template('regulations/layers/{}.html'.format(key))
            for key in ('table', 'note', 'code', 'subscript', 'dash',
                        'footnote', 'superscript')}

    def render_table(self, table, data_type=None):
        max_width = 0
        for header_row in table['header']:
            width = sum(cell['colspan'] for cell in header_row)
            max_width = max(max_width, width)

        #  Just in case a row is longer than the header
        row_max = max(len(row) for row in table['rows'])
        max_width = max(max_width, row_max)

        #  Now pad rows if needed
        for row in table['rows']:
            row.extend([''] * (max_width - len(row)))

        context = Context(table)
        #   Remove new lines so that they don't get escaped on display
        return self.tpls['table'].render(context).replace('\n', '')

    def render_fence(self, fence, data_type=None):
        """Fenced paragraphs are formatted separately, offset from the rest of
        the text. They have an associated "type" which further specifies their
        format"""
        _type = fence.get('type')
        lines = fence.get('lines', [])
        strip_nl = True
        if _type == 'note':
            lines = [l.replace('Note:', '').replace('Notes:', '')
                     for l in lines]
            lines = [l for l in lines if l.strip()]
            tpl = self.tpls['note']
        else:   # Generic "code"/ preformatted
            strip_nl = False
            tpl = self.tpls['code']

        rendered = tpl.render(Context({'lines': lines, 'type': _type}))
        if strip_nl:
            rendered = rendered.replace('\n', '')
        return rendered

    def render_replacement(self, data, data_type):
        """Several of the formatted data types are simple string replacements.
        Implement them all here"""
        context = Context(data)
        return self.tpls[data_type].render(context).replace('\n', '')

    def replacements_for(self, original, data):
        for data_type in self.DATA_TYPES:
            processor = getattr(self, 'render_' + data_type,
                                self.render_replacement)
            key = data_type + '_data'
            if key in data:
                yield processor(data[key], data_type)



from django.template import loader, Context

from regulations.generator.layers.base import InlineLayer
from regulations.generator.section_url import SectionUrl


class InternalCitationLayer(InlineLayer):
    shorthand = 'internal'
    data_source = 'internal-citations'

    def __init__(self, layer):
        self.layer = layer
        self.sectional = False
        self.version = None
        self.rev_urls = SectionUrl()
        self.rendered = {}

    def render_url(self, label, text,
                   template_name='regulations/layers/internal_citation.html'):

        key = (tuple(label), text, template_name)
        if key not in self.rendered:
            url = self.rev_urls.fetch(label, self.version, self.sectional)
            c = Context({'citation': {
                'url': url, 'label': text,
                'label_id': self.rev_urls.view_label_id(label, self.version)}})
            template = loader.get_template(template_name)
            self.rendered[key] = template.render(c).strip('\n')
        return self.rendered[key]

    def replacement_for(self, original, data):
        return self.render_url(data['citation'], original)



from .base import *

DEBUG = False
ALLOWED_HOSTS = ['locahost', '127.0.0.1']

# In production, this needs to be set to the directory where the static files
# should end up:
# STATIC_ROOT = ''

# In production, set this to what makes sense for you.
CACHE_MIDDLEWARE_SECONDS = 600

try:
    from local_settings import *
except ImportError:
    pass



from mock import Mock, patch
from django.conf import settings
from unittest import TestCase

from regulations.generator.layers.interpretations import InterpretationsLayer


class InterpretationsLayerTest(TestCase):
    def setUp(self):
        if not settings.configured:
            settings.configure(API_BASE='')

    @patch('regulations.generator.layers.interpretations.generator')
    @patch('regulations.generator.layers.interpretations.SectionUrl')
    def test_apply_layer_extra_fields(self, su, generator):
        su.return_value.interp.return_value = '200-Subpart-Interp'
        layer = {
            "200-2-b-3-i": [{
                'reference': '200-2-b-3-i-Interp',
                "text": "Some contents are here"
            }],
        }
        il = InterpretationsLayer(layer)
        il.partial_view = Mock()
        il.partial_view.return_value.content = 'content'

        self.assertEqual(il.apply_layer('200-2-b-3-i'), ('interp', {
            'for_markup_id': '200-2-b-3-i',
            'for_label': '2(b)(3)(i)',
            'interps': [{
                'label_id': '200-2-b-3-i-Interp',
                'markup': 'content',
                'section_id': '200-Subpart-Interp'}]
        }))

    @patch('regulations.generator.layers.interpretations.generator')
    @patch('regulations.generator.layers.interpretations.SectionUrl')
    def test_apply_layer_section(self, su, generator):
        layer = {
            "200-2": [{
                "reference": "200-2-Interp",
                "text": "Some contents are here"
            }],
        }
        il = InterpretationsLayer(layer)
        il.partial_view = Mock()
        il.partial_view.return_value.content = 'content'

        self.assertEqual('2', il.apply_layer('200-2')[1]['for_label'])

    @patch('regulations.generator.layers.interpretations.generator')
    @patch('regulations.generator.layers.interpretations.SectionUrl')
    def test_apply_layer_multiple_matches(self, su, generator):
        layer = {
            "200-2": [{
                "reference": "200-2-Interp",
                "text": "Some contents are here"
            }, {
                "reference": "200-2_3-Interp",
                "text": "Some more contents are here"
            }],
        }
        il = InterpretationsLayer(layer)
        il.partial_view = Mock()
        il.partial_view.return_value.content = 'content'

        _, data = il.apply_layer('200-2')
        labels = [interp['label_id'] for interp in data['interps']]
        self.assertEqual(labels, ['200-2-Interp', '200-2_3-Interp'])

    @patch('regulations.generator.layers.interpretations.generator')
    @patch('regulations.generator.layers.interpretations.SectionUrl')
    def test_apply_layer_appendix(self, su, piv):
        layer = {
            "200-Q-5": [{
                "reference": "200-Q-5-Interp",
                "text": "Some contents are here"
            }],
        }
        il = InterpretationsLayer(layer)
        il.partial_view = Mock()
        il.partial_view.return_value.content = 'content'

        self.assertEqual('Appendix Q-5',
                         il.apply_layer('200-Q-5')[1]['for_label'])

    @patch('regulations.generator.layers.interpretations.generator')
    @patch('regulations.generator.layers.interpretations.SectionUrl')
    def test_apply_layer_section_different(self, su, generator):
        layer = {
            "200-2-a": [{
                "reference": "200-2-a-Interp",
                "text": "Some contents are here"
            }],
            "200-2-b": [{
                "reference": "200-2-a-Interp",
                "text": "Some contents are here"
            }],
        }
        il = InterpretationsLayer(layer)
        il.partial_view = Mock()
        il.partial_view.return_value.content = 'content'

        _, result = il.apply_layer('200-2-a')
        self.assertEqual('2(a)', result['for_label'])

        _, result = il.apply_layer('200-2-b')
        self.assertEqual('2(b)', result['for_label'])

    @patch('regulations.generator.layers.interpretations.generator')
    @patch('regulations.generator.layers.interpretations.SectionUrl')
    def test_apply_layer_cache(self, secturl, generator):
        il = InterpretationsLayer({
            '1234-56-a': [{'reference': '1234-56-a-Interp'}]}, version='vvvv')
        il.root_interp_label = '1234-56-Interp'
        il.partial_view = Mock()
        il.partial_view.return_value.content = 'content'
        il.apply_layer('1234-56')
        self.assertFalse(generator.generator.get_tree_paragraph.called)

        il.apply_layer('1234-56-a')
        self.assertTrue(generator.generator.get_tree_paragraph.called)
        args = generator.generator.get_tree_paragraph.call_args[0]
        # Note that this is grabbing the section's interps
        self.assertEqual('1234-56-Interp', args[0])
        self.assertEqual('vvvv', args[1])

    @patch('regulations.generator.layers.interpretations.views')
    def test_preprocess_root(self, views):
        node = {'text': 'tttt', 'children': [], 'node_type': 'regtext',
                'label': ['1234', '56', 'a']}
        il = InterpretationsLayer({})
        il.preprocess_root(node)
        self.assertEqual(il.root_interp_label, '1234-56-a-Interp')



import os
import json
import time
import logging

import celery
from django.conf import settings
from django.core.cache import caches
from django.shortcuts import redirect
from django.http import Http404, JsonResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods
from django.template.response import TemplateResponse
from django.utils.crypto import get_random_string
from django.views.generic.base import View
import requests

from regulations import tasks
from regulations import docket
from regulations.views.preamble import (
    common_context, CommentState, generate_html_tree, first_preamble_section,
    notice_data)

logger = logging.getLogger(__name__)

# TODO: Expire preview URL at commenting deadline
PREVIEW_EXPIRATION_SECONDS = 60 * 60 * 24 * 90


def upload_proxy(request):
    """Create a random key name and a pair of temporary PUT and GET URLS to
    permit attachment uploads and previews from the browser.
    """
    filename = request.GET['name']
    size = int(request.GET['size'])
    valid, message = validate_attachment(filename, size)
    if not valid:
        logger.error(message)
        return JsonResponse({'message': message}, status=400)
    key, put_url = tasks.SignedUrl.generate(
        method='put_object',
        params={
            'ContentLength': size,
            'ContentType': request.GET.get('type', 'application/octet-stream'),
            'Metadata': {'name': filename},
        },
    )
    disposition = 'attachment; filename="{}"'.format(filename)
    _, get_url = tasks.SignedUrl.generate(
        key=key,
        params={
            'ResponseExpires': time.time() + PREVIEW_EXPIRATION_SECONDS,
            'ResponseContentDisposition': disposition,
        },
    )
    return JsonResponse({
        'urls': {'get': get_url, 'put': put_url},
        'key': key,
    })


@csrf_exempt
@require_http_methods(['POST'])
def preview_comment(request):
    """Convert a comment to PDF, upload the result to S3, and return a signed
    URL to GET the PDF.
    """
    body = json.loads(request.body.decode('utf-8'))
    sections = body.get('assembled_comment', [])
    html = tasks.json_to_html(sections, mark_as_draft=True)
    key = '/'.join([settings.ATTACHMENT_PREVIEW_PREFIX, get_random_string(50)])
    document_number = tasks.get_document_number(sections)
    content_disposition = tasks.generate_content_disposition(
        document_number, draft=True)
    with tasks.html_to_pdf(html) as pdf:
        tasks.s3_client.put_object(
            Body=pdf,
            ContentType='application/pdf',
            ContentDisposition=content_disposition,
            Bucket=settings.ATTACHMENT_BUCKET,
            Key=key,
        )
    _, url = tasks.SignedUrl.generate(key=key)
    return JsonResponse({'url': url})


regs_gov_fmt = 'https://www.regulations.gov/#!documentDetail;D={document}'


class SubmitCommentView(View):

    def get(self, request, doc_number):
        preamble, _, _ = notice_data(doc_number)
        section = first_preamble_section(preamble)
        if section is None:
            raise Http404
        return redirect(
            'chrome_preamble', paragraphs='/'.join(section['label']))

    def post(self, request, doc_number):
        form_data = {
            key: value for key, value in request.POST.items()
            if key != 'comments'
        }
        comments = json.loads(request.POST.get('comments', '[]'))

        context = common_context(doc_number)

        if context['meta']['comment_state'] != CommentState.OPEN:
            raise Http404("Cannot comment on {}".format(doc_number))

        context.update(generate_html_tree(context['preamble'], request,
                                          id_prefix=[doc_number, 'preamble']))
        context['comment_mode'] = 'write'
        context.update({'message': '', 'metadata_url': ''})

        valid, context['message'] = self.validate(comments, form_data)
        context['regs_gov_url'] = regs_gov_fmt.format(
            document=settings.COMMENT_DOCUMENT_ID)

        # Catch any errors related to enqueueing comment submission. Because
        # this step can fail for many reasons (e.g. no connection to broker,
        # broker fails to write, etc.), catch `Exception`.
        try:
            _, context['metadata_url'] = self.enqueue(comments, form_data)
        except Exception as exc:
            logger.exception(exc)

        template = 'regulations/comment-confirm-chrome.html'
        return TemplateResponse(request=request, template=template,
                                context=context)

    def validate(self, comments, form_data):
        valid, message = docket.sanitize_fields(form_data)
        if not valid:
            logger.error(message)
            return valid, message

        files = tasks.extract_files(comments)
        # Account for the main comment itself submitted as an attachment
        if len(files) > settings.MAX_ATTACHMENT_COUNT - 1:
            message = "Too many attachments"
            logger.error(message)
            return False, message

        return True, ''

    def enqueue(self, comments, form_data):
        metadata_url = tasks.SignedUrl.generate()
        chain = celery.chain(
            tasks.submit_comment.s(comments, form_data, metadata_url),
            tasks.publish_tracking_number.s(metadata_url=metadata_url),
        )
        chain.delay()
        return metadata_url


@require_http_methods(['GET', 'HEAD'])
def get_federal_agencies(request):
    return lookup_regulations_gov(field='gov_agency',
                                  dependentOnValue='Federal')


@require_http_methods(['GET', 'HEAD'])
def get_gov_agency_types(request):
    return lookup_regulations_gov(field='gov_agency_type')


def lookup_regulations_gov(*args, **kwargs):
    """ GET lookup values from regulations.gov. Use a cache """
    cache = caches['regs_gov_cache']
    cache_key = make_cache_key(**kwargs)
    response = cache.get(cache_key)

    if response is None:
        logger.debug("Looking up in regs.gov")
        response = requests.get(
            settings.REGS_GOV_API_LOOKUP_URL,
            params=kwargs,
            headers={'X-Api-Key': settings.REGS_GOV_API_KEY}
        )
        if response.status_code == requests.codes.ok:
            response = JsonResponse(response.json()['list'], safe=False)
            cache.set(cache_key, response)
        else:
            logger.error("Failed to lookup regulations.gov: {}",
                         response.status_code, response.text)
            response.raise_for_status()
    return response


def validate_attachment(filename, size):
    if size <= 0 or size > settings.ATTACHMENT_MAX_SIZE:
        return False, "Invalid attachment size"
    _, ext = os.path.splitext(filename)
    if ext[1:].lower() not in settings.VALID_ATTACHMENT_EXTENSIONS:
        return False, "Invalid attachment type"
    return True, ""


def make_cache_key(*args, **kwargs):
    """ Make a cache key of the form key1:value1:key2:value2.
        Sort the keys to ensure repeatability
    """
    return ":".join((key + ":" + str(kwargs[key]) for key in sorted(kwargs)))



from datetime import date

from django.conf import settings
from django.utils.decorators import decorator_from_middleware_with_args
from django.views.decorators.cache import cache_page
from django.middleware.cache import CacheMiddleware


lt_cache = cache_page(settings.CACHES['eregs_longterm_cache']['TIMEOUT'],
                      cache='eregs_longterm_cache')


class DailyCacheMiddleware(CacheMiddleware):
    """Like the cache middleware, but always expires at midnight"""
    @property
    def key_prefix(self):
        return date.today().isoformat() + '/' + (self.__key_prefix or '')

    @key_prefix.setter
    def key_prefix(self, value):
        self.__key_prefix = value


daily_cache = decorator_from_middleware_with_args(DailyCacheMiddleware)(
    cache_timeout=settings.CACHES['eregs_longterm_cache']['TIMEOUT'],
    cache_alias='eregs_longterm_cache')



import logging

import six
import requests

from django.conf import settings
from django.core.cache import caches

logger = logging.getLogger(__name__)

cache = caches['regs_gov_cache']


def fetch(url, **kwargs):
    kwargs['headers'] = kwargs.get(
        'headers', {'X-Api-Key': settings.REGS_GOV_API_KEY})
    kwargs['timeout'] = kwargs.get('timeout', 10)   # timeout after 10 seconds
    response = requests.get(url, **kwargs)
    if response.status_code != requests.codes.ok:
        response.raise_for_status()
    return response.json()


def get_document_fields(document_id):
    """ Retrieve the field list from the document metadata.
        Use a cache as the data hardly ever changes.
        We ignore the 'general_comment' field as it has special treatment.
    """
    if getattr(settings, 'REGS_GOV_API_MOCK', None):
        return {}
    fields = cache.get(document_id)
    if fields is not None:
        return fields

    metadata = fetch(settings.REGS_GOV_API_URL, params={'D': document_id})
    fields = {
        field['attributeName']: field
        for field in metadata['fieldList']
        if field['attributeName'] != 'general_comment'
    }

    add_picklist_options(fields)
    add_combo_options(fields)

    cache.set(document_id, fields)
    return fields


def safe_get_document_fields(document_id):
    """Like get_document_fields, but returns None on error"""
    try:
        return get_document_fields(document_id)
    except requests.exceptions.RequestException as e:
        logger.warning("Error getting data from regs.gov: %s", e)
        return None


def add_picklist_options(fields):
    """Augment list fields with options. Adds a list of options to each field
    of type "picklist".
    """
    for name, field in fields.items():
        if field['uiControl'] == 'picklist':
            data = fetch(field['lookupUrl'])
            field['options'] = data['list']


def add_combo_options(fields):
    """Augment combo fields with dependency-contingent options. Adds a dict
    mapping contingent values to lists of options to each field of type
    "combo".
    """
    for name, field in fields.items():
        if field['uiControl'] == 'combo':
            field['options'] = {}
            for option in fields[field['dependsOn']]['options']:
                data = fetch(field['lookupUrl'] + option['value'])
                field['options'][option['value']] = data.get('list', [])


def sanitize_fields(body):
    """ Validate fields against the document metadata.
        Remove any extra fields that are not in the document metadata.
        Special treatment for 'assembled_comment' - allow to pass through
        as it holds the text form of the entire submission
    """
    document_fields = safe_get_document_fields(settings.COMMENT_DOCUMENT_ID)
    if document_fields is None:
        return True, ""

    for name, field in six.iteritems(document_fields):
        if field['required'] and name not in body:
            return False, "Field {} is required".format(name)
        if name in body and len(body[name]) > field['maxLength']:
            return False, "Field {} exceeds maximum length of {}".format(
                name, field['maxLength'])

    # Remove extra fields if any, other than 'assembled_comment'
    extra_fields = [field for field in body if field not in document_fields]
    for field in extra_fields:
        if field != 'assembled_comment':
            del body[field]
    return True, ""



from datetime import date

from django.http import HttpResponse
from django.test import TestCase
from django.test.client import RequestFactory
from django.utils.decorators import decorator_from_middleware_with_args
from mock import Mock, patch

from regulations import url_caches


class UrlCachesTests(TestCase):
    @patch('regulations.url_caches.date')
    def test_daily_cache(self, patched_date):
        """Cache should be consistent within a day but not between days"""
        fn = Mock(return_value=HttpResponse('response'))
        request = RequestFactory().get('a-path')

        mock_caches = {'example': {
            'BACKEND': 'django.core.cache.backends.locmem.LocMemCache'}}
        with self.settings(CACHES=mock_caches):
            daily_cache = decorator_from_middleware_with_args(
                url_caches.DailyCacheMiddleware)(cache_alias='example')
            wrapped_fn = daily_cache(fn)

        patched_date.today.return_value = date(2010, 10, 10)
        self.assertEqual(fn.call_count, 0)
        wrapped_fn(request)
        self.assertEqual(fn.call_count, 1)
        wrapped_fn(request)
        self.assertEqual(fn.call_count, 1)

        patched_date.today.return_value = date(2010, 10, 11)
        wrapped_fn(request)
        self.assertEqual(fn.call_count, 2)
        wrapped_fn(request)
        self.assertEqual(fn.call_count, 2)



from unittest import TestCase

from regulations.generator.layers.footnotes import FootnotesLayer


class FootnotesLayerTest(TestCase):
    def test_single_note(self):
        layer = {
            "555-220-p1": [
                {
                    "footnote_data": {
                        "note": "Some notes",
                        "ref": "1"
                    },
                    "locations": [
                        0
                    ],
                    "text": "[^1]Some actual text"
                }
            ]
        }
        fl = FootnotesLayer(layer)
        self.assertEqual(fl.apply_layer('555-220'), ('footnotes',
                         [{'ref': '1',
                           'note': 'Some notes'}]))

    def test_sorted_multiple_notes(self):
        layer = {
            "555-220-p1": [
                {
                    "footnote_data": {
                        "note": "Third notes",
                        "ref": "3"
                    },
                    "locations": [
                        0
                    ],
                    "text": "[^3]Some actual text"
                },
                {
                    "footnote_data": {
                        "note": "First notes",
                        "ref": "1"
                    },
                    "locations": [
                        0
                    ],
                    "text": "[^1]Some actual text"
                },
                {
                    "footnote_data": {
                        "note": "Second notes",
                        "ref": "2"
                    },
                    "locations": [
                        0
                    ],
                    "text": "[^2]Some actual text"
                }
            ]
        }
        fl = FootnotesLayer(layer)
        self.assertEqual(fl.apply_layer('555-220'), ('footnotes',
                         [
                             {'ref': '1', 'note': 'First notes'},
                             {'ref': '2', 'note': 'Second notes'},
                             {'ref': '3', 'note': 'Third notes'},
                         ]))

    def test_multiple_children(self):
        layer = {
            "555-220-p1": [
                {
                    "footnote_data": {
                        "note": "First notes",
                        "ref": "1"
                    },
                    "locations": [
                        0
                    ],
                    "text": "[^1]Some actual text"
                }
            ],
            "555-220-p2": [
                {
                    "footnote_data": {
                        "note": "Second notes",
                        "ref": "2"
                    },
                    "locations": [
                        0
                    ],
                    "text": "[^1]Some actual text"
                }
            ]
        }
        fl = FootnotesLayer(layer)
        self.assertEqual(fl.apply_layer('555-220'), ('footnotes',
                         [
                             {'ref': '1', 'note': 'First notes'},
                             {'ref': '2', 'note': 'Second notes'},
                         ]))



from BeautifulSoup import BeautifulSoup
import requests
import threading
import sys
import time

from six.moves.urllib_parse import urlparse


class EregsCache():

    @staticmethod
    def write(msg):
        sys.stdout.write(msg + "\n")

    @staticmethod
    def write_error(msg):
        sys.stderr.write(msg + "\n")

    def access_url(self, *args):

        try:
            req = requests.get(args[0])

            status_txt = "Cached"
            if req.status_code != 200:
                status_txt = "Failed"

            msg = "{0} (status {1}): {2}".format(
                status_txt, str(req.status_code), req.url)
            self.write(msg)
        except Exception as err:
            self.write_error("Failed: " + args[0])
            self.write_error(str(err))

    def get_main_reg_list(self):
        try:
            regulations_links = []
            html = requests.get(self.full_url).text
            soup = BeautifulSoup(html)
            reg_list = soup.find("ul", {"class": "reg-list"})

            for link in reg_list.findAll("li"):
                regulations_links.append(link.find("a")["href"])

            return regulations_links

        except Exception as err:
            self.write_error("Main Page Failed To Load")
            self.write_error(str(err))

    # figure out the partial path depending on root location of eregs
    def get_partial_url(self,  href):

        href = href.split("/")

        if self.full_url != self.base_url:
            href[1] += "/partial"
        else:
            href[0] += "/partial"

        href = '/'.join(href)

        href = href.split("#")[0]

        return self.base_url + href

    def __init__(self, eregs_url, regulations=None):
        """ kick off calling regulations website for caching purposes

            Keyword arguments:
            eregs_url -- full web url to main regulations page
            regulations -- (optional) list of regulation paths to be processed
        """

        url = urlparse(eregs_url)
        self.base_url = url.scheme + "://" + url.netloc
        self.full_url = eregs_url

        self.write("Base Url:" + self.base_url)
        self.write("eRegs location:" + self.full_url)

        try:
            if not regulations:
                regulations_links = self.get_main_reg_list()
            else:
                regulations_links = regulations.split(",")

            for link in regulations_links:
                self.write("Getting NAV links from " + self.base_url + link)

                reg_nav = requests.get(self.base_url+link).text
                soup = BeautifulSoup(reg_nav)
                reg_soup = soup.findAll("a")

                for a in reg_soup:
                    if 'data-section-id' in a:
                        partial = self.get_partial_url(a["href"])

                        thread_count = len(threading.enumerate())

                        # process 5 web pages at time. Some servers hate a lot
                        # of requests at once.  slow it down so the threads
                        # can catch up if there are too many threads being
                        # processed, end process and exit out
                        if thread_count <= 5:
                            threading.Thread(target=self.access_url,
                                             args=(partial, )).start()
                        elif thread_count >= 12:
                            self.write("URLs currently at " + str(thread_count)
                                       + ". Server too slow")
                            self.write("Shutting down")
                            raise Exception("Thread Count Too High")
                        else:
                            self.write("Currently Processing "
                                       + str(thread_count) + " Urls")
                            self.write("Waiting...")
                            time.sleep(thread_count * 2)

                # let the threads catch up before doing the next batch
                if len(threading.enumerate()) > 1:
                    time.sleep(len(threading.enumerate()) * 2)

        except Exception as err:
            self.write_error(str(err))


if __name__ == "__main__":
    regulations_arg = None
    if len(sys.argv) > 3:
        regulations_arg = sys.argv[3]

    EregsCache(sys.argv[1], regulations_arg)



from unittest import TestCase

from django.test import RequestFactory
from mock import patch

from regulations.views import redirect


class ViewsRedirectTest(TestCase):
    @patch('regulations.views.redirect.ApiReader')
    def test_redirect_by_date(self, ApiReader):
        ApiReader.return_value.regversions.return_value = {'versions': [
            {'by_date': '2000-01-01', 'version': 'aaa'},
            {'by_date': '2005-05-05', 'version': 'bbb'},
            {'by_date': '2010-06-07', 'version': 'ccc'},
        ]}

        with patch('regulations.views.redirect.handle_generic_404') as handle:
            redirect.redirect_by_date(None, '8888', '1999', '10', '10')
            self.assertTrue(handle.called)

            handle.reset_mock()
            redirect.redirect_by_date(None, '8888', '', '', '')
            self.assertTrue(handle.called)

        response = redirect.redirect_by_date(None, '8888', '2000', '01', '01')
        self.assertEqual(302, response.status_code)
        self.assertTrue('aaa' in response['Location'])
        response = redirect.redirect_by_date(None, '8888', '2006', '06', '06')
        self.assertEqual(302, response.status_code)
        self.assertTrue('bbb' in response['Location'])
        response = redirect.redirect_by_date(None, '8888', '2010', '06', '08')
        self.assertEqual(302, response.status_code)
        self.assertTrue('ccc' in response['Location'])

    @patch('regulations.views.redirect.redirect_by_date')
    def test_redirect_by_date_get(self, redirect_by_date):
        request = RequestFactory().get('?year=2222&month=11&day=20')
        redirect.redirect_by_date_get(request, 'lablab')
        self.assertTrue(redirect_by_date.called)
        self.assertEqual(('lablab', '2222', '11', '20'),
                         redirect_by_date.call_args[0][1:])

        request = RequestFactory().get('?year=-2222&month=-11&day=-20')
        redirect.redirect_by_date_get(request, 'lablab')
        self.assertTrue(redirect_by_date.called)
        self.assertEqual(('lablab', '2222', '11', '20'),
                         redirect_by_date.call_args[0][1:])

        request = RequestFactory().get('?year=123&month=1&day=2')
        redirect_by_date.reset_mock()
        redirect.redirect_by_date_get(request, 'lablab')
        self.assertTrue(redirect_by_date.called)
        self.assertEqual(('lablab', '0123', '01', '02'),
                         redirect_by_date.call_args[0][1:])

        request = RequestFactory().get('?year=22&month=1&day=2')
        redirect_by_date.reset_mock()
        redirect.redirect_by_date_get(request, 'lablab')
        self.assertTrue(redirect_by_date.called)
        self.assertEqual(('lablab', '2022', '01', '02'),
                         redirect_by_date.call_args[0][1:])

        with patch('regulations.views.redirect.handle_generic_404') as handle:
            request = RequestFactory().get('?year=bad&month=data&day=here')
            redirect.redirect_by_date_get(request, 'lablab')
            self.assertTrue(handle.called)

        with patch('regulations.views.redirect.handle_generic_404') as handle:
            request = RequestFactory().get('')
            redirect.redirect_by_date_get(request, 'lablab')
            self.assertTrue(handle.called)

    def test_diff_redirect_bad_version(self):
        request = RequestFactory().get('?new_version=A+Bad+Version')
        response = redirect.diff_redirect(request, 'lablab', 'verver')
        self.assertEqual(404, response.status_code)

    @patch('regulations.views.redirect.fetch_grouped_history')
    def test_diff_redirect_order(self, fgh):
        fgh.return_value = [
            {'notices': [{'document_number': '3'}, {'document_number': '2'}]},
            {'notices': [{'document_number': '1'}]}]
        request = RequestFactory().get('?new_version=3')
        response = redirect.diff_redirect(request, '1111-22', '1')
        self.assertTrue('diff/1111-22/1/3' in response['Location'])
        self.assertTrue('from_version=1' in response['Location'])
        response = redirect.diff_redirect(request, '1111-22', '2')
        self.assertTrue('diff/1111-22/2/3' in response['Location'])
        self.assertTrue('from_version=2' in response['Location'])

        request = RequestFactory().get('?new_version=2')
        response = redirect.diff_redirect(request, '1111-22', '1')
        self.assertTrue('diff/1111-22/1/2' in response['Location'])
        self.assertTrue('from_version=1' in response['Location'])
        response = redirect.diff_redirect(request, '1111-22', '3')
        self.assertTrue('diff/1111-22/2/3' in response['Location'])
        self.assertTrue('from_version=3' in response['Location'])

        request = RequestFactory().get('?new_version=1')
        response = redirect.diff_redirect(request, '1111-22', '2')
        self.assertTrue('diff/1111-22/1/2' in response['Location'])
        self.assertTrue('from_version=2' in response['Location'])
        response = redirect.diff_redirect(request, '1111-22', '3')
        self.assertTrue('diff/1111-22/1/3' in response['Location'])
        self.assertTrue('from_version=3' in response['Location'])



from itertools import takewhile

from cached_property import cached_property

from regulations.generator.layers.tree_builder import make_label_sortable


def sort_regtext_label(label):
    """ Make a regtext label sortable """
    sortable = [make_label_sortable(l)[0] for l in label]
    if len(sortable) > 4:
        sortable[4] = make_label_sortable(sortable[4], roman=True)[0]
    return sortable


class Label(object):
    """Rather than flip back and forth between list and string versions of
    node labels, wrap both here. Also provide properties for inspecting the
    label (e.g. to determine if it represents an interpretation, if another
    label is a child of this one"""
    def __init__(self, id=None, parts=None):
        if id is None:
            id = '-'.join(parts)
        if parts is None:
            parts = id.split('-')

        self._id = id
        self._parts = parts

    id = property(lambda self: self._id)
    parts = property(lambda self: self._parts)

    @cached_property
    def is_interp(self):
        return 'Interp' in self.parts

    @cached_property
    def is_interp_root(self):
        return self.parts and self.parts[-1] == 'Interp'

    @cached_property
    def prefix(self):
        """Interpretations have an added complication:
            1005-2-Interp-2 is a child of 1005-Interp"""
        return tuple(takewhile(lambda k: k != 'Interp', self.parts))

    @cached_property
    def sort_key(self):
        if not self.is_interp:
            return tuple(sort_regtext_label(self.parts))
        else:
            suffix = self.parts[len(self.prefix) + 1:]
            prefix_sortable = sort_regtext_label(self.prefix)
            suffix_sortable = [make_label_sortable(l)[0] for l in suffix]
            if len(suffix_sortable) > 1:
                suffix_sortable[1] = make_label_sortable(
                    suffix_sortable[1], roman=True)[0]
            return tuple(prefix_sortable + suffix_sortable)

    def __contains__(self, other):
        """Is the `other` label a child of this one?"""
        if isinstance(other, Label):
            prefix_match = other.prefix[:len(self.prefix)] == self.prefix
            lexical_match = other.parts[:len(self.parts)] == self.parts

            child_of_interp_root = (
                self.is_interp_root and other.is_interp and prefix_match)
            lexical_child = (
                lexical_match and other.is_interp == self.is_interp)
            return child_of_interp_root or lexical_child
        else:
            return False

    def __lt__(self, other):
        return self.sort_key < other.sort_key

    def __eq__(self, other):
        return isinstance(other, Label) and self.id == other.id



from unittest import TestCase
from mock import patch
from django.test import RequestFactory

from regulations.views.chrome_breakaway import ChromeSXSView


class ChromeSXSViewTests(TestCase):

    @patch('regulations.views.chrome_breakaway.api_reader')
    @patch('regulations.views.chrome_breakaway.ChromeSXSView.content')
    def test_get(self, content, api_reader):
        content.return_value = ''
        api_reader.ApiReader.return_value.layer.return_value = {
            '204': [''],
        }
        request = RequestFactory().get(
            '/fake-path/204-2/2013-1?from_version=2014-2')

        view = ChromeSXSView.as_view(
            template_name='regulations/breakaway-chrome.html')
        response = view(
            request, label_id='204-2', from_version='2014-2',
            notice_id='2013-1')

        self.assertEqual(response.context_data['reg_part'], '204')



from datetime import datetime, timedelta
import random
from unittest import TestCase

from mock import patch

from regulations.views import universal_landing as universal


class UniversalLandingTest(TestCase):
    """ Tests for the view that drives the main (universal) landing page. """

    def test_filter_future_amendments(self):
        versions = []
        futures = []

        today = datetime.today()
        for i in range(1, 5):
            future_date = today + timedelta(days=i)
            v = {'by_date': future_date}
            versions.append(v)
            futures.append(v)

        for i in range(1, 3):
            past_date = today - timedelta(days=i)
            versions.append({'by_date': past_date})

        random.shuffle(versions)

        self.assertEqual(len(versions), 6)
        filtered = universal.filter_future_amendments(versions)
        self.assertEqual(len(filtered), 4)
        self.assertEqual(futures, filtered)

    def test_get_regulations_list_sort(self):
        """Verify that part numbers are sorted numerically rather than
        lexicographically"""
        version_info = [{'version': 'v', 'by_date': datetime(2001, 1, 1)}]
        versions = {'1': version_info, '2': version_info, '100': version_info}
        with patch('regulations.views.utils.regulation_meta'):
            with patch('regulations.views.utils.first_section'):
                results = universal.get_regulations_list(versions)
                self.assertEqual(['1', '2', '100'],
                                 [r['part'] for r in results])



from unittest import TestCase

from regulations.generator import notices


class NoticesTest(TestCase):
    def test_find_label_in_sxs_found(self):
        sxs_list = [
            {'label': '204-1', 'children': []},
            {'label': '204-2', 'children': [{
                'label': '204-2-a',
                'children': [
                    {'label': '204-3', 'children': [], 'paragraphs': ['x']}],
                'paragraphs': ['abc']}]}]

        s = notices.find_label_in_sxs(sxs_list, '204-2-a')
        self.assertEqual('204-2-a', s['label'])
        self.assertEqual(['abc'], s['paragraphs'])

        s = notices.find_label_in_sxs(sxs_list, '204-3')
        self.assertEqual('204-3', s['label'])
        self.assertEqual(['x'], s['paragraphs'])

        sxs_list = [
            {'labels': ['204-1'], 'children': []},
            {'labels': ['204-2'], 'children': [{
                'labels': ['204-2-a', '204-2-b'],
                'children': [
                    {'labels': ['204-3'], 'children': [],
                     'paragraphs': ['x']}],
                'paragraphs': ['abc']}]}]

        s = notices.find_label_in_sxs(sxs_list, '204-2-b')
        self.assertEqual(['204-2-a', '204-2-b'], s['labels'])
        self.assertEqual(['abc'], s['paragraphs'])

        s = notices.find_label_in_sxs(sxs_list, '204-3')
        self.assertEqual(['204-3'], s['labels'])
        self.assertEqual(['x'], s['paragraphs'])

    def test_find_label_in_sxs_top_no_label(self):
        sxs_list = [
            {'title': 'Awesome, SXS title here', 'children': [
                {'label': '204-3', 'children': [], 'paragraphs': ['x']}],
                'paragraphs': ['abc']}]

        s = notices.find_label_in_sxs(sxs_list, '204-3')
        self.assertEqual('204-3', s['label'])
        self.assertEqual(['x'], s['paragraphs'])

        sxs_list = [
            {'title': 'Awesome, SXS title here', 'children': [
                {'labels': ['204-3'], 'children': [], 'paragraphs': ['x']}],
                'paragraphs': ['abc']}]

        s = notices.find_label_in_sxs(sxs_list, '204-3')
        self.assertEqual(['204-3'], s['labels'])
        self.assertEqual(['x'], s['paragraphs'])

    def test_find_label_in_sxs_page(self):
        sxs_list = [
            {'labels': ['204-3'], 'page': 1234, 'paragraphs': ['a'],
             'children': [
                {'labels': ['204-3-a'], 'page': 1234, 'paragraphs': ['b'],
                 'children': []}]},
            {'labels': ['204-3'], 'page': 3456, 'paragraphs': ['c'],
             'children': [
                {'labels': ['204-3-a'], 'page': 3457, 'paragraphs': ['d'],
                 'children': []},
                {'labels': ['204-3-a'], 'page': 3460, 'paragraphs': ['e'],
                 'children': []}]}]

        s = notices.find_label_in_sxs(sxs_list, '204-3')
        self.assertEqual(['a'], s['paragraphs'])
        s = notices.find_label_in_sxs(sxs_list, '204-3', 1234)
        self.assertEqual(['a'], s['paragraphs'])
        s = notices.find_label_in_sxs(sxs_list, '204-3', 9999)
        self.assertEqual(['a'], s['paragraphs'])
        s = notices.find_label_in_sxs(sxs_list, '204-3', 3456)
        self.assertEqual(['c'], s['paragraphs'])

        s = notices.find_label_in_sxs(sxs_list, '204-3-a')
        self.assertEqual(['b'], s['paragraphs'])
        s = notices.find_label_in_sxs(sxs_list, '204-3-a', 1234)
        self.assertEqual(['b'], s['paragraphs'])
        s = notices.find_label_in_sxs(sxs_list, '204-3-a', 9999)
        self.assertEqual(['b'], s['paragraphs'])
        s = notices.find_label_in_sxs(sxs_list, '204-3-a', 3457)
        self.assertEqual(['d'], s['paragraphs'])
        s = notices.find_label_in_sxs(sxs_list, '204-3-a', 3460)
        self.assertEqual(['e'], s['paragraphs'])

    def test_non_empty_sxs(self):
        sxs = {'label': '204-2-a', 'children': [], 'paragraphs': ['abc']}
        self.assertTrue(notices.non_empty_sxs(sxs))

    def test_non_empty_sxs_no_paragraph(self):
        sxs = {'label': '204-2-a', 'children': [], 'paragraphs': []}
        self.assertFalse(notices.non_empty_sxs(sxs))

    def test_non_empty_sxs_has_children(self):
        sxs = {
            'label': '204-2-a',
            'children': [{'title': 'abc'}],
            'paragraphs': []}
        self.assertTrue(notices.non_empty_sxs(sxs))

    def test_find_label_in_sxs_not_found(self):
        sxs_list = [
            {'label': '204-1', 'children': []},
            {'label': '204-2', 'children': [{
                'label': '204-2-a',
                'children': []}]}]

        s = notices.find_label_in_sxs(sxs_list, '202-a')
        self.assertEqual(None, s)

    def test_filter_children(self):
        sxs = {'children': [
            {'label': '204-a', 'paragraphs': ['me']},
            {'paragraphs': ['abcd']}]}
        filtered = notices.filter_labeled_children(sxs)
        self.assertEqual(filtered, [{'paragraphs': ['abcd']}])

    def test_filter_children_no_candidates(self):
        sxs = {'children': [
            {'label': '204-a', 'paragraphs': ['me']},
            {'label': '204-b', 'paragraphs': ['abcd']}]}
        filtered = notices.filter_labeled_children(sxs)
        self.assertEqual(filtered, [])

    def test_add_depths(self):
        sxs = {
            'label': '204-2',
            'children': [{
                'label': '204-2-a',
                'children': [],
                'paragraphs': ['abc']}]}
        notices.add_depths(sxs, 3)
        depth_sxs = {
            'label': '204-2',
            'depth': 3,
            'children': [{
                'depth': 4,
                'label': '204-2-a',
                'children': [],
                'paragraphs': ['abc']}]}
        self.assertEqual(depth_sxs, sxs)



from unittest import TestCase

from mock import patch

from regulations.generator.api_reader import ApiReader


class ClientTest(TestCase):
    @patch('regulations.generator.api_reader.api_client')
    def test_regulation(self, api_client):
        to_return = {'example': 0, 'label': ['204'], 'children': []}
        api_client.ApiClient.return_value.get.return_value = to_return
        reader = ApiReader()
        self.assertEqual(to_return,
                         reader.regulation("label-here", "date-here"))
        self.assertTrue(api_client.ApiClient.return_value.get.called)
        param = api_client.ApiClient.return_value.get.call_args[0][0]
        self.assertTrue('label-here' in param)
        self.assertTrue('date-here' in param)

    @patch('regulations.generator.api_reader.api_client')
    def test_layer(self, api_client):
        to_return = {'example': 1}
        api_client.ApiClient.return_value.get.return_value = to_return
        reader = ApiReader()
        self.assertEqual(
            to_return,
            reader.layer("layer-here", "cfr", "label-here", "version-here"))
        get = api_client.ApiClient.return_value.get
        self.assertEqual(1, get.call_count)
        param = api_client.ApiClient.return_value.get.call_args[0][0]
        self.assertIn('layer-here/cfr/version-here/label', param)
        self.assertNotIn('label-here', param)   # only grabs the root, "label"

        #   Cache
        self.assertEqual(
            to_return,
            reader.layer("layer-here", "cfr", "label-abc", "version-here"))
        self.assertEqual(1, get.call_count)

        self.assertEqual(
            to_return,
            reader.layer("layer-here", "cfr", "lablab", "version-here"))
        self.assertEqual(2, get.call_count)
        param = get.call_args[0][0]
        self.assertIn('layer-here/cfr/version-here/lablab', param)

        self.assertEqual(
            to_return,
            reader.layer("layer-here", "preamble", "lablab"))
        self.assertEqual(3, get.call_count)
        param = get.call_args[0][0]
        self.assertIn('layer-here/preamble/lablab', param)

    @patch('regulations.generator.api_reader.api_client')
    def test_notices(self, api_client):
        to_return = {'example': 1}
        api_client.ApiClient.return_value.get.return_value = to_return
        reader = ApiReader()
        self.assertEqual(to_return, reader.notices())
        get = api_client.ApiClient.return_value.get
        self.assertTrue(get.called)

        self.assertEqual(to_return, reader.notices(part='p'))
        self.assertTrue(get.called)
        self.assertEqual({'part': 'p'}, get.call_args[0][1])

    @patch('regulations.generator.api_reader.api_client')
    def test_regversion(self, api_client):
        to_return = {}
        api_client.ApiClient.return_value.get.return_value = to_return
        reader = ApiReader()
        self.assertEqual(to_return, reader.regversions('765'))
        get = api_client.ApiClient.return_value.get
        self.assertTrue(get.called)
        param = get.call_args[0][0]
        self.assertTrue('765' in param)

    @patch('regulations.generator.api_reader.api_client')
    def test_notice(self, api_client):
        to_return = {'example': 1}
        api_client.ApiClient.return_value.get.return_value = to_return
        reader = ApiReader()
        self.assertEqual(to_return, reader.notice("doc"))
        get = api_client.ApiClient.return_value.get
        self.assertTrue(get.called)
        param = get.call_args[0][0]
        self.assertTrue('doc' in param)

    @patch('regulations.generator.api_reader.api_client')
    def test_diff(self, api_client):
        to_return = {'example': 1}
        api_client.ApiClient.return_value.get.return_value = to_return
        reader = ApiReader()
        self.assertEqual(to_return, reader.diff("204", "old", "new"))

        get = api_client.ApiClient.return_value.get
        self.assertTrue(get.called)
        param = get.call_args[0][0]
        self.assertTrue('204' in param)
        self.assertTrue('old' in param)
        self.assertTrue('new' in param)

    @patch('regulations.generator.api_reader.api_client')
    def test_reg_cache(self, api_client):
        child = {
            'text': 'child',
            'node_type': 'interp',
            'children': [],
            'label': ['923', 'a', 'Interp'],
            'title': 'Some title'
        }
        child2 = {
            'text': 'child2',
            'node_type': 'interp',
            'children': [],
            'label': ['923', 'Interp', '1']
        }
        child3 = {
            'text': 'child',
            'node_type': 'interp',
            'children': [],
            'label': ['923', 'b', 'Interp'],
        }
        to_return = {
            'text': 'parent',
            'node_type': 'interp',
            'label': ['923', 'Interp'],
            'children': [child, child2, child3]
        }
        get = api_client.ApiClient.return_value.get
        get.return_value = to_return
        reader = ApiReader()

        reader.regulation('923-Interp', 'ver')
        reader.regulation('923-Interp', 'ver')
        reader.regulation('923-a-Interp', 'ver')
        self.assertEqual(1, get.call_count)

        get.return_value = child2
        reader.regulation('923-Interp-1', 'ver')
        self.assertEqual(2, get.call_count)

        get.return_value = child3
        reader.regulation('923-b-Interp', 'ver')
        self.assertEqual(3, get.call_count)

        child = {
            'text': 'child',
            'children': [],
            'label': ['923', '1', 'a']
        }
        to_return = {
            'text': 'parent',
            'label': ['923', '1'],
            'children': [child]
        }
        get.reset_mock()
        get.return_value = to_return
        reader.regulation('923-1', 'ver')
        reader.regulation('923-1', 'ver')
        reader.regulation('923-1-a', 'ver')
        get = api_client.ApiClient.return_value.get
        self.assertEqual(2, get.call_count)

    @patch('regulations.generator.api_reader.api_client')
    def test_cache_mutability(self, api_client):
        to_return = {
            'text': 'parent',
            'label': ['1024'],
            'children': []
        }
        get = api_client.ApiClient.return_value.get
        get.return_value = to_return
        reader = ApiReader()

        result = reader.regulation('1024', 'ver')
        self.assertEqual(to_return, result)
        child = {
            'text': 'child',
            'children': [],
            'label': ['1024', 'a']
        }
        result['children'] = [child]

        second = reader.regulation('1024', 'ver')
        self.assertEqual(1, get.call_count)
        self.assertEqual(second, {'text': 'parent', 'label': ['1024'],
                                  'children': []})



from unittest import TestCase
from regulations.generator.layers import layers_applier
from regulations.generator.layers import location_replace


class LayersApplierTest(TestCase):

    def test_enqueue(self):
        applier = layers_applier.LayersApplier()
        element = ('abcd', 'ABCD', [])
        applier.enqueue(element)

        priority, retrieved = applier.queue.get()
        self.assertEquals(priority, -4)
        self.assertEquals(retrieved, element)

    def test_list_enqueue(self):
        applier = layers_applier.LayersApplier()
        elements = [('abcd', 'ABCD', []), ('efghi', 'EFG', [12])]
        applier.enqueue_from_list(elements)

        priority, retrieved = applier.queue.get()
        self.assertEquals(priority, -5)
        self.assertEqual(elements[1], retrieved)

        priority, retrieved = applier.queue.get()
        self.assertEquals(priority, -4)
        self.assertEqual(elements[0], retrieved)

    def test_replace_all(self):
        applier = layers_applier.LayersApplier()
        applier.text = ('Prefix test <a \nhref="url" data="test">link '
                        '\ntest</a> postfix text')
        applier.replace_all('test', 'linksecondword')

        replaced = ('Prefix linksecondword <a \nhref="url" data="test">link '
                    '\nlinksecondword</a> postfix text')
        self.assertEquals(applier.text, replaced)

    def test_find_all_offsets(self):
        pattern = 'ABCD'
        text = 'The grey fox ABCD jumped over the fence ABCD'
        offsets = location_replace.LocationReplace.find_all_offsets(pattern,
                                                                    text)
        self.assertEquals(offsets, [(13, 17), (40, 44)])

    def test_find_offsets_no_pattern(self):
        pattern = 'ABCD'
        text = 'The grey fox jumped over the fence'
        offsets = location_replace.LocationReplace.find_all_offsets(pattern,
                                                                    text)
        self.assertEquals(offsets, [])

    def test_replace_at_offset(self):
        pattern = 'ABCD'
        text = 'The grey fox ABCD jumped over the fence ABCD'

        first_offset = location_replace.LocationReplace.find_all_offsets(
            pattern, text)[0]
        result = location_replace.LocationReplace.replace_at_offset(
            first_offset, 'giraffe', text)

        self.assertEquals(result,
                          'The grey fox giraffe jumped over the fence ABCD')

    def test_replace_at(self):
        text = 'The grey fox ABCD jumped ABCD over the fence ABCD'

        applier = layers_applier.LayersApplier()
        applier.text = text
        applier.replace_at('ABCD', '<a>ABCD</a>', [0, 2])

        self.assertEquals(
            applier.text,
            'The grey fox <a>ABCD</a> jumped ABCD over the fence <a>ABCD</a>')

    def test_update_offsets(self):
        lr = location_replace.LocationReplace()
        lr.offset_starter = 5

        pattern = 'ABCD'
        text = ('The grey <a href="link">ABCD</a> jumped over the ABCD fence '
                'on a ABCD day')

        lr.update_offsets(pattern, text)
        self.assertEquals(lr.counter, 0)
        self.assertEqual(lr.offset_starter, 5)
        self.assertEqual(lr.offset_counters, [5, 6, 7])
        self.assertEqual(list(lr.offsets.keys()), [5, 6, 7])
        self.assertEqual(lr.offsets[5], (24, 28))

    def test_update_offset_starter(self):
        lr = location_replace.LocationReplace()
        lr.offset_counters = [5, 6, 7]
        lr.update_offset_starter()

        self.assertEqual(lr.offset_starter, 8)

    def test_replace_at_case_sensitive(self):
        original = 'state'
        replacement = '<a href="link_url">state</a>'
        locations = [0, 1, 2]

        applier = layers_applier.LayersApplier()
        applier.text = "<em>(6)</em> <dfn> Under state law. </dfn> State law."
        applier.replace_at(original, replacement, locations)

        result = (u"<em>(6)</em> <dfn> Under <a href=\"link_url\">state</a> "
                  u"law. </dfn> State law.")
        self.assertEquals(applier.text, result)

    def test_replace_no_original(self):
        original = 'federal'
        replacement = '<a href="link_url">state</a>'
        locations = [0, 1, 2]

        applier = layers_applier.LayersApplier()
        applier.text = "<em>(6)</em> <dfn> Under state law. </dfn> State law."
        applier.replace_at(original, replacement, locations)

        result = "<em>(6)</em> <dfn> Under state law. </dfn> State law."
        self.assertEquals(applier.text, result)

    def test_replace_skip_location(self):
        original = 'state'
        replacement = '<a href="link_url">state</a>'
        locations = [0, 2]

        applier = layers_applier.LayersApplier()
        applier.text = ("<em>(6)</em> <dfn> Under state law. </dfn> state "
                        "law. <dfn> state liability. </dfn>")
        applier.replace_at(original, replacement, locations)

        result = ("<em>(6)</em> <dfn> Under <a href=\"link_url\">state</a> "
                  "law. </dfn> state law. <dfn> <a href=\"link_url\">state"
                  "</a> liability. </dfn>")
        self.assertEquals(applier.text, result)



from unittest import TestCase
from django.core.urlresolvers import reverse


class UrlTests(TestCase):
    def test_about(self):
        r = reverse('regulations.views.about.about')
        self.assertEqual(r, '/about')

    def test_chrome_section_url(self):
        r = reverse('chrome_section_view', args=('201-2', '2012-1123'))
        self.assertEqual(r, '/201-2/2012-1123')

        r = reverse(
            'chrome_section_view', args=('201-2', '2012-1123_20121011'))
        self.assertEqual(r, '/201-2/2012-1123_20121011')

    def test_sxs_url(self):
        r = reverse('chrome_sxs_view', args=('201-2-g', '2011-1738'))
        self.assertEqual(r, '/sxs/201-2-g/2011-1738')

        r = reverse('chrome_sxs_view', args=('201-2-g', '2011-1738_20110232'))
        self.assertEqual(r, '/sxs/201-2-g/2011-1738_20110232')

    def test_diff_url(self):
        r = reverse(
            'chrome_section_diff_view',
            args=('201-2', '2011-1738', '2012-22345'))
        self.assertEqual(r, '/diff/201-2/2011-1738/2012-22345')

        r = reverse(
            'chrome_section_diff_view',
            args=('201-2', '2011-1738_20121011', '2012-22345_20131022'))
        self.assertEqual(
            r, '/diff/201-2/2011-1738_20121011/2012-22345_20131022')



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from collections import namedtuple
from copy import deepcopy
from datetime import date
from enum import Enum

import logging

from django.http import Http404
from django.conf import settings
from django.shortcuts import redirect
from django.core.urlresolvers import reverse
from django.template.response import TemplateResponse
from django.views.generic.base import View

from fr_notices import navigation
from regulations import docket
from regulations.generator.api_reader import ApiReader
from regulations.generator.generator import LayerCreator
from regulations.generator.html_builder import (
    CFRChangeHTMLBuilder, PreambleHTMLBuilder)
from regulations.generator.layers.utils import (
    convert_to_python, is_contained_in)
from regulations.views import utils
from regulations.views import chrome
from regulations.views.diff import Versions, get_appliers


logger = logging.getLogger(__name__)


class CommentState(Enum):
    NO_COMMENT, PREPUB, OPEN, CLOSED = range(4)


def find_subtree(root, label_parts):
    """Given a nested tree and a label to look for, find the associated node
    in the tree. Note that, unlike regulations, preamble trees _always_ encode
    their exact position in the tree."""
    cursor = root
    while cursor and cursor['label'] != label_parts:
        next_cursor = None
        for child in cursor['children']:
            if child['label'] == label_parts[:len(child['label'])]:
                next_cursor = child
        cursor = next_cursor
    return cursor


def generate_html_tree(subtree, request, id_prefix=None):
    """Use the HTMLBuilder to generate a version of this subtree with
    appropriate markup. Currently, includes no layers"""
    layer_creator = LayerCreator()
    doc_id = '-'.join(subtree['label'])
    layer_creator.add_layers(utils.layer_names(request), 'preamble',
                             doc_id, sectional=True)
    builder = PreambleHTMLBuilder(
        *layer_creator.get_appliers(),
        id_prefix=id_prefix,
        index_prefix=[0, subtree.get('lft')]
    )
    builder.tree = subtree
    builder.generate_html()

    return {'node': builder.tree,
            'markup_page_type': 'reg-section'}


def merge_cfr_changes(doc_number, notice):
    """We started with a mock version of these changes which were stored as a
    setting, CFR_CHANGES. Until we remove that completely, merge those values
    with real data from the notice structure"""
    mock_notice = getattr(settings, 'CFR_CHANGES', {}).get(doc_number) or {}

    versions = dict(mock_notice.get('versions', {}))        # copy
    versions.update(notice.get('versions', {}))

    amendments = list(mock_notice.get('amendments', []))    # copy
    amendments.extend(notice.get('amendments', []))

    return versions, amendments


def notice_data(doc_number):
    preamble = ApiReader().preamble(doc_number.replace('-', '_'))
    if preamble is None:
        raise Http404

    notice = ApiReader().notice(doc_number.replace('_', '-')) or {}

    fields = (
        "amendments",
        "cfr_parts",
        "cfr_title",
        "comment_doc_id",
        "comments_close",
        "dockets",
        "document_number",
        "effective_on",
        "footnotes",
        "fr_citation",
        "fr_url",
        "fr_volume",
        "meta",
        "primary_agency",
        "primary_docket",
        "publication_date",
        "regulation_id_numbers",
        "section_by_section",
        "supporting_documents",
        "title",
        "versions"
    )

    meta = {}
    for field in fields:
        if field in notice:
            meta[field] = convert_to_python(deepcopy(notice[field]))

    # If there's no metadata, fall back to getting it from settings:
    if not meta:
        meta = getattr(settings, 'PREAMBLE_INTRO', {}).get(doc_number, {}).get(
            'meta', {})
        meta = convert_to_python(deepcopy(meta))

    today = date.today()
    if 'comments_close' in meta and 'publication_date' in meta:
        close_date = meta['comments_close'].date()
        publish_date = meta['publication_date'].date()
        if today < publish_date:
            meta['comment_state'] = CommentState.PREPUB
        elif today <= close_date:
            meta['comment_state'] = CommentState.OPEN
            meta['days_remaining'] = 1 + (close_date - today).days
        else:
            meta['comment_state'] = CommentState.CLOSED
    else:
        meta['comment_state'] = CommentState.NO_COMMENT

    # We don't pass along cfr_ref information in a super useful format, yet.
    # Construct one here:
    if 'cfr_refs' not in meta and 'cfr_title' in meta and 'cfr_parts' in meta:
        meta['cfr_refs'] = [{"title": meta['cfr_title'],
                             "parts": meta['cfr_parts']}]

    return preamble, meta, notice


def first_preamble_section(preamble):
    return next(
        (
            node for node in preamble['children']
            if not node['label'][-1].startswith('p')
        ),
        None,
    )


def common_context(doc_number):
    """All of the "preamble" views share common context, such as preamble
    data, toc info, etc. This function retrieves that data and returns the
    results as a dict. This may throw a 404"""
    preamble, meta, notice = notice_data(doc_number)
    preamble_toc = navigation.make_preamble_nav(preamble['children'])
    versions, amendments = merge_cfr_changes(doc_number, notice)
    cfr_toc = navigation.make_cfr_change_nav(doc_number, versions, amendments)

    return {
        'cfr_change_toc': cfr_toc,
        'doc_number': doc_number,
        'meta': meta,
        'notice': notice,
        'preamble': preamble,
        'preamble_toc': preamble_toc,
    }


class PreambleView(View):
    """Displays either a notice preamble (or a subtree of that preamble). If
    using AJAX or specifically requesting, generate only the preamble markup;
    otherwise wrap it in the appropriate "chrome" """
    def get(self, request, *args, **kwargs):
        label_parts = kwargs.get('paragraphs', '').split('/')
        doc_number = label_parts[0]
        context = common_context(doc_number)

        # Redirect to first section on top-level preamble
        if len(label_parts) == 1:
            section = first_preamble_section(context['preamble'])
            if not section:
                raise Http404
            return redirect(
                'chrome_preamble', paragraphs='/'.join(section['label']))

        subtree = find_subtree(context['preamble'], label_parts)
        if subtree is None:
            raise Http404

        sub_context = generate_html_tree(subtree, request,
                                         id_prefix=[doc_number, 'preamble'])
        template = sub_context['node']['template_name']
        nav = navigation.footer(
            context['preamble_toc'],
            context['cfr_change_toc'],
            full_id=sub_context['node']['full_id'],
        )
        sub_context['meta'] = context['meta']

        context.update({
            'sub_context': sub_context,
            'sub_template': template,
            'full_id': sub_context['node']['full_id'],
            'section_label': sub_context['node']['human_label'],
            'type': 'preamble',
            'navigation': nav,
        })

        if not request.is_ajax() and request.GET.get('partial') != 'true':
            template = 'regulations/preamble-chrome.html'
        else:
            template = 'regulations/preamble-partial.html'
        return TemplateResponse(request=request, template=template,
                                context=context)


class ChromePreambleSearchView(chrome.ChromeSearchView):
    template_name = 'regulations/chrome-preamble-search.html'

    def get_context_data(self, **kwargs):
        label_parts = kwargs.get('label_id', '').split('/')
        doc_number = label_parts[0]
        context = common_context(doc_number)
        context['doc_type'] = 'preamble'
        context['label_id'] = kwargs.get('label_id', '')

        subtree = find_subtree(context['preamble'], label_parts)
        if subtree is None:
            raise Http404

        context.update(generate_html_tree(context['preamble'], self.request,
                                          id_prefix=[doc_number, 'preamble']))
        self.add_main_content(context)
        return context


class PrepareCommentView(View):
    def get(self, request, doc_number):
        context = common_context(doc_number)

        if context['meta']['comment_state'] != CommentState.OPEN:
            raise Http404("Cannot comment on {}".format(doc_number))

        context.update(generate_html_tree(context['preamble'], request,
                                          id_prefix=[doc_number, 'preamble']))
        context['comment_mode'] = 'write'
        context['comment_fields'] = docket.safe_get_document_fields(
            settings.COMMENT_DOCUMENT_ID)
        template = 'regulations/comment-review-chrome.html'

        return TemplateResponse(request=request, template=template,
                                context=context)


SubpartInfo = namedtuple('SubpartInfo', ['letter', 'title', 'urls', 'idx'])


# @todo - this shares a lot of code w/ PreambleView. Can we merge them?
class CFRChangesView(View):
    def get(self, request, doc_number, section):
        context = common_context(doc_number)
        full_id = '{}-cfr-{}'.format(doc_number, section)

        versions, amendments = merge_cfr_changes(doc_number, context['notice'])
        label_parts = section.split('-')

        if len(label_parts) == 1:
            sub_context = self.authorities_context(
                amendments, cfr_part=section)
            section_label = 'Authority'
        else:
            toc_position = next(
                idx for idx, nav in enumerate(context['cfr_change_toc'])
                if nav.section_id == full_id)
            sub_context = self.regtext_changes_context(
                amendments,
                versions,
                doc_number=doc_number,
                label_id=section,
                toc_position=toc_position,
            )
            section_label = sub_context['tree']['human_label']
        sub_context['meta'] = context['meta']

        context.update({
            'sub_context': sub_context,
            'sub_template': 'regulations/cfr_changes.html',
            'full_id': full_id,
            'section_label': section_label,
            'type': 'cfr',
            'navigation': navigation.footer(
                context['preamble_toc'],
                context['cfr_change_toc'],
                full_id
            ),
        })

        if not request.is_ajax() and request.GET.get('partial') != 'true':
            template = 'regulations/preamble-chrome.html'
        else:
            template = 'regulations/preamble-partial.html'
        return TemplateResponse(request=request, template=template,
                                context=context)

    @staticmethod
    def authorities_context(amendments, cfr_part):
        """What authorities information is relevant to this CFR part?"""
        relevant = [amd for amd in amendments
                    if amd.get('cfr_part') == cfr_part and 'authority' in amd]
        return {'instructions': [a['instruction'] for a in relevant],
                'authorities': [a['authority'] for a in relevant]}

    @classmethod
    def regtext_changes_context(cls, amendments, version_info, label_id,
                                doc_number, toc_position):
        """Generate diffs for the changed section"""
        cfr_part = label_id.split('-')[0]
        relevant = []
        for amd in amendments:
            changes = amd.get('changes', [])
            keys = {change[0] for change in changes}
            if any(is_contained_in(key, label_id) for key in keys):
                relevant.append(amd)

        versions = Versions(version_info[cfr_part]['left'],
                            version_info[cfr_part]['right'])
        left_tree = ApiReader().regulation(label_id, versions.older)
        appliers = get_appliers(label_id, versions)

        builder = CFRChangeHTMLBuilder(
            *appliers, id_prefix=[str(doc_number), 'cfr'],
            index_prefix=[1, toc_position]
        )
        builder.tree = left_tree or {}
        builder.generate_html()

        return {
            'instructions': [a['instruction'] for a in relevant],
            'subparts': list(cls.subpart_changes(doc_number, relevant,
                                                 label_id)),
            'tree': builder.tree,
        }

    @staticmethod
    def subpart_changes(doc_number, amendments, label_id):
        """Meta data about additional subparts; we'll pass this through to the
        template"""
        subpart_changes = [change['node']
                           for amd in amendments
                           for key, change_list in amd.get('changes', [])
                           for change in change_list
                           if 'Subpart' in key]
        for change in subpart_changes:
            urls = [
                reverse('cfr_changes',
                        kwargs={'doc_number': doc_number, 'section': section})
                for section in change['child_labels']]
            yield SubpartInfo(letter=change['label'][-1],
                              title=change['title'],
                              urls=urls,
                              idx=change['child_labels'].index(label_id))



import unittest

from selenium.webdriver.support.ui import WebDriverWait

from regulations.uitests.base_test import BaseTest


class InterpTest(BaseTest, unittest.TestCase):

    job_name = 'Interp test'

    def test_interps(self):
        self.driver.get(self.test_url + '/1005-2/2012-12121')
        html = self.driver.find_element_by_tag_name('html')
        WebDriverWait(self.driver, 30).until(
            lambda driver: 'selenium-start' in html.get_attribute('class'))

        interp_dropdown = self.driver.find_element_by_xpath(
            '//*[@id="1005-2-h"]/section')

        # interp should know who it belongs to
        self.assertEquals(interp_dropdown.get_attribute('data-interp-for'),
                          '1005-2-h')

        # interp section should know what is in it
        self.assertEqual(interp_dropdown.get_attribute('data-interp-id'),
                         '1005-18-a')

        # should have the appropriate header
        self.assertIn('OFFICIAL INTERPRETATION TO 2(h)',
                      interp_dropdown.text)

        self.driver.execute_script(
            'p10052h = document.getElementById("1005-2-h").offsetTop')
        self.driver.execute_script('window.scrollTo(0, p10052h)')

        # body should be hidden
        interp_text = self.driver.find_element_by_xpath(
            '//*[@id="1005-2-h"]/section/section')
        interp_dropdown.click()

        WebDriverWait(self.driver, 120).until(
            lambda driver: driver.find_element_by_css_selector(
                '.inline-interpretation.open'))

        # header should update
        self.assertIn('HIDE', interp_dropdown.text)

        # should contain the appropriate reg section
        WebDriverWait(self.driver, 10).until(
            lambda driver: interp_text.text)
        self.assertIn('clicked. A finances centripetally curiousest '
                      'stronghold cemeteries', interp_text.text)

        self.driver.find_element_by_xpath(
            '//*[@id="1005-2-h"]/section/header/a').click()

        WebDriverWait(self.driver, 10).until(
            lambda driver: driver.find_element_by_css_selector(
                '.inline-interpretation:not(.open)'))

        # header should reflect close
        self.assertIn('SHOW', interp_dropdown.text)



class LocationReplace(object):
    """ Applies location based layers to XML nodes. We use XML so that we only
    take into account the original text when we're doing a replacement. """
    def __init__(self):
        self.counter = 0
        self.offset_starter = 0
        self.offset_counters = None
        self.offsets = None

    @staticmethod
    def find_all_offsets(pattern, text, offset=0):
        """Don't use regular expressions as they are a tad slow"""
        matches = []
        pattern_len = len(pattern)
        next_match = text.find(pattern)
        while next_match != -1:
            matches.append((next_match + offset,
                            next_match + pattern_len + offset))
            next_match = text.find(pattern, next_match + 1)
        return matches

    @staticmethod
    def replace_at_offset(offset, replacement, text):
        return text[:offset[0]] + replacement + text[offset[1]:]

    def update_offsets(self, original, text):
        """ Offsets change everytime we replace the text, since we add more
        characters. Update the offsets. """
        list_offsets = []
        lt = text.find('<')
        gt = -1
        while lt != -1:
            subtext = text[gt+1: lt]
            list_offsets.extend(LocationReplace.find_all_offsets(
                original, subtext, gt + 1))
            gt = text.find('>', lt)
            lt = text.find('<', gt)
        list_offsets.extend(LocationReplace.find_all_offsets(
            original, text[gt+1:], gt + 1))

        self.offset_counters = list(range(self.offset_starter,
                                    self.offset_starter + len(list_offsets)))
        self.offsets = dict(zip(self.offset_counters, list_offsets))

    def update_offset_starter(self):
        """ As we're navigating the XML node, we need to keep track of how many
        offsets we've already seen. """
        if len(self.offset_counters) > 0:
            self.offset_starter = self.offset_counters[-1] + 1

    def location_replace_text(self, text, original, replacement, locations):
        """Given plain text, do replacements"""
        self.update_offsets(original, text)

        text_segments = []
        relevant_locations = sorted(self.offsets.keys())
        relevant_locations = [l for l in relevant_locations if l in locations]
        text_begin = 0
        for location in relevant_locations:
            start, end = self.offsets[location]
            # unrelated text
            text_segments.append(text[text_begin:start])
            # s/original/replacement
            text_segments.append(replacement)
            text_begin = end
        # tail of unrelated text
        text_segments.append(text[text_begin:])

        # offset_starter is shared between segments of xml nodes (in
        # location_replace, below)
        if original not in replacement:
            self.offset_starter += len(locations)

        self.update_offset_starter()
        return "".join(text_segments)

    def location_replace(self, xml_node, original, replacement, locations):
        """ For the xml_node, replace the locations instances of orginal with
        replacement.
        @todo: This doesn't appear to be used anymore?"""

        if xml_node.text:
            xml_node.text = self.location_replace_text(
                xml_node.text, original, replacement, locations)

        for c in xml_node.getchildren():
            self.location_replace(c, original, replacement, locations)

        if xml_node.tail:
            xml_node.tail = self.location_replace_text(
                xml_node.tail, original, replacement, locations)



from unittest import TestCase

from mock import patch

from regulations.generator import generator
from regulations.generator.layers.layers_applier import InlineLayersApplier
from regulations.generator.layers.layers_applier import ParagraphLayersApplier
from regulations.generator.layers.layers_applier\
    import SearchReplaceLayersApplier


class GeneratorTest(TestCase):

    @patch('regulations.generator.generator.api_reader')
    def test_get_regulation_extra_fields(self, api_reader):
        reg = {
            'text': '', 'children': [], 'label': ['8675'],
            'title': 'Contains no part info'
        }
        api_reader.ApiReader.return_value.regulation.return_value = reg

        r = generator.get_regulation('8675', 'ver')
        self.assertFalse('title_clean' in r)
        self.assertFalse('reg_letter' in r)

        reg['title'] = 'part 8675 - Some title'
        r = generator.get_regulation('8675', 'ver')
        self.assertTrue('title_clean' in r)
        self.assertEqual('Some title', r['title_clean'])
        self.assertFalse('reg_letter' in r)

        del reg['title_clean']
        reg['title'] = 'part 8675 - Some title (RegUlation RR)'
        r = generator.get_regulation('8675', 'ver')
        self.assertTrue('title_clean' in r)
        self.assertEqual('Some title', r['title_clean'])
        self.assertTrue('reg_letter' in r)
        self.assertEqual('RR', r['reg_letter'])

    @patch('regulations.generator.generator.api_reader')
    def test_get_tree_paragraph(self, api_reader):
        node = {'some': 'text'}
        api_reader.ApiReader.return_value.regulation.return_value = node

        p = generator.get_tree_paragraph('some-id', 'some-version')
        self.assertEqual(node, p)
        self.assertEqual(
            ('some-id', 'some-version'),
            api_reader.ApiReader.return_value.regulation.call_args[0])

    @patch('regulations.generator.generator.api_reader')
    def test_get_diff_json(self, api_reader):
        diff = {'some': 'diff'}
        api_reader.ApiReader.return_value.diff.return_value = diff
        d = generator.get_diff_json('204', 'old', 'new')
        self.assertEqual(diff, d)
        self.assertEqual(
            ('204', 'old', 'new'),
            api_reader.ApiReader.return_value.diff.call_args[0])

    @patch('regulations.generator.generator.api_reader')
    def test_get_notice(self, api_reader):
        notice = {'some': 'notice'}
        api_reader.ApiReader.return_value.notice.return_value = notice
        n = generator.get_notice('204-1234')
        self.assertEqual(notice, n)
        self.assertEqual(
            ('204-1234',),
            api_reader.ApiReader.return_value.notice.call_args[0])

    @patch('regulations.generator.generator.get_diff_json')
    def test_get_diff_applier(self, get_diff_json):
        diff = {'some': 'diff'}
        get_diff_json.return_value = diff
        da = generator.get_diff_applier('204', 'old', 'new')
        self.assertEqual(da.diff, diff)
        self.assertEqual(
            ('204', 'old', 'new'),
            get_diff_json.call_args[0])

    @patch('regulations.generator.generator.get_diff_json')
    def test_get_diff_applier_allows_empty(self, get_diff_json):
        diff = {}
        get_diff_json.return_value = diff
        da = generator.get_diff_applier('204', 'old', 'new')
        self.assertEqual(da.diff, diff)
        self.assertEqual(
            ('204', 'old', 'new'),
            get_diff_json.call_args[0])

    def test_layercreator_getappliers(self):
        creator = generator.LayerCreator()
        appliers = creator.get_appliers()
        self.assertEquals(len(appliers), 3)

        i_applier, p_applier, s_applier = appliers

        self.assertTrue(isinstance(i_applier, InlineLayersApplier))
        self.assertTrue(isinstance(p_applier, ParagraphLayersApplier))
        self.assertTrue(isinstance(s_applier, SearchReplaceLayersApplier))

    @patch('regulations.generator.generator.LayerCreator.get_layer_json')
    def test_add_layers(self, get_layer_json):
        get_layer_json.return_value = {'layer': 'layer'}

        creator = generator.LayerCreator()
        creator.add_layers(
            ['meta', 'graphics', 'internal'], 'cfr', '205',
            sectional=True, version='verver')
        i, p, s = creator.get_appliers()
        self.assertEquals(len(p.layers), 1)
        self.assertEquals(len(i.layers), 1)
        self.assertEquals(len(s.layers), 1)

        internal_citation_layer = i.layers['internal']
        self.assertTrue(internal_citation_layer.sectional)
        self.assertEquals(internal_citation_layer.version, 'verver')



from django_nose.runner import NoseTestSuiteRunner


class DatabaselessTestRunner(NoseTestSuiteRunner):
    """ A test suite runner that does not setup and tear down a database. """

    def setup_databases(self):
        pass

    def teardown_databases(self, *args):
        pass



import json
import mock

from django.test import SimpleTestCase, override_settings

from regulations.views import comment


@override_settings(
    ATTACHMENT_BUCKET='test-bucket',
    ATTACHMENT_ACCESS_KEY_ID='test-access-key',
    ATTACHMENT_SECRET_ACCESS_KEY='test-secret-key',
    ATTACHMENT_MAX_SIZE=42,
)
class TestUploadProxy(SimpleTestCase):

    @mock.patch('time.time')
    @mock.patch('regulations.tasks.s3_client')
    @mock.patch('regulations.tasks.get_random_string')
    def test_get_url(self, get_random, mock_client, mock_time):
        generate_presigned = mock_client.generate_presigned_url
        generate_presigned.side_effect = ['first-url', 'second-url']
        get_random.return_value = 'not-so-random'
        mock_time.return_value = 123
        resp = self.client.get('/comments/attachment?size=42&name=foo.pdf')
        self.assertEqual(resp.status_code, 200)
        body = json.loads(resp.content.decode())
        generate_presigned.assert_any_call(
            ClientMethod='put_object',
            Params={
                'ContentLength': 42,
                'ContentType': 'application/octet-stream',
                'Bucket': 'test-bucket',
                'Key': get_random.return_value,
                'Metadata': {'name': 'foo.pdf'}
            },
        )
        generate_presigned.assert_any_call(
            ClientMethod='get_object',
            Params={
                'ResponseExpires': 123 + comment.PREVIEW_EXPIRATION_SECONDS,
                'ResponseContentDisposition': 'attachment; filename="foo.pdf"',
                'Bucket': 'test-bucket',
                'Key': get_random.return_value,
            },
        )
        self.assertEqual(body['key'], get_random.return_value)
        self.assertEqual(body['urls']['put'], 'first-url')
        self.assertEqual(body['urls']['get'], 'second-url')

    def test_get_url_empty(self):
        resp = self.client.get('/comments/attachment?size=0&name=foo.pdf')
        self.assertEqual(resp.status_code, 400)
        body = json.loads(resp.content.decode())
        self.assertEqual(body['message'], 'Invalid attachment size')

    def test_get_url_over_limit(self):
        resp = self.client.get('/comments/attachment?size=43&name=foo.pdf')
        self.assertEqual(resp.status_code, 400)
        body = json.loads(resp.content.decode())
        self.assertEqual(body['message'], 'Invalid attachment size')

    def test_get_url_invalid_extension(self):
        resp = self.client.get('/comments/attachment?size=42&name=foo.exe')
        self.assertEqual(resp.status_code, 400)
        body = json.loads(resp.content.decode())
        self.assertEqual(body['message'], 'Invalid attachment type')
